/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **/
/*                                                                        */
/*      Analog - Digital Demonstration Program for C language             */
/*      Copyright (c) 1986 - 1993 Strawberry Tree Incorporated            */
/*                                                                        */
/* ---------------------------------------------------------------------- */
/* NOTE:                                                                  */ 
/*      Small Memory Model:                                               */
/*          Use CCALL() and link with the module MS_CALL.OBJ            */
/*      Medium Memory Model:                                              */
/*          Use CCALL() with the below function prototype and             */
/*          link with MS_CALL.OBJ                                         */
/*          Void CCALL(char far *, float far *, unsigned far *);          */
/*      Compact Memory Model:                                             */
/*          Use CCALL() with the below function prototype and           */
/*          link with MS_CALL.OBJ                                         */
/*          Void CCALL(char near *, float near *, unsigned near *);     */
/*      Large/Huge Memory Model:                                          */
/*          Use CCALL() and link with MS_CALL.OBJ                         */
/*                                                                        */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **/

#include <stdio.h>

#define   ANACNT  16                      /* Number of analog channels */
#define   DIGCNT  16                      /* Number of digital channels */

unsigned  digital[DIGCNT];                /* Digital channel array */
float     analog[ANACNT];                 /* Analog channel array */
int       i;
int       range;                          /* Gain range for card  */


/*----------------------------------------------------------------
	initial  -  Initialize A-D board.  Check and report any
		        errors to user.
*/

void    initial()
{
    /** Read CALIB.DAT file and calibrate the analog inputs **/

    CCALL( "Fn" , analog, digital );         /* Call A-D driver */

    if (digital[0] == 0 && digital[2] == 0)  {
	printf("\nDriver, ADRIVE.COM or GDRV.COM, not installed; ");
	 printf(" or analog card not installed.\n");
	exit(1);
    }
    if (digital[0] == 0 && digital[2] != 0)  {
	printf("\nNo analog card selected. BRD SEL switch set to 0.\n");
	exit(1);
    }
    if (digital[0] > digital[6])  {
	printf("\nCalibration numbers are not correct.\n");
	exit(1);
    }
    if (digital[0] > ANACNT || digital[2] > DIGCNT)  {
	printf("\nToo many channels installed.\n");
	printf("Change size of ANACNT and/or DIGCNT in heading\n");
	exit(1);
    }

    /** Set analog channel resolution */
    /** For ACJr, Mini-16 & ACPC-X-X cards, res = 18 gives low noise mode */
    digital[0] = 18;
    CCALL( "a" , analog, digital );

    /** Find card type  **/
    CCALL( "B" , analog, digital );

	/* Set range for ACJr, Mini-16, or ACPC-X-X is 10V  */
    range = 17;
	/** Flash-12 Mod 1 range is +/-10V **/
    /** Flash-12 Mod 2 is always +/- 5V **/
/*    if (digital[1] == 513) (range = 7); */
	/** WorkMate range is +/-5V **/
/*    if (digital[1] == 641) (range = 8); */
    for (i = 0; i < ANACNT; ++i) digital[i] = 17;
    CCALL( "r" , analog, digital );

    /** Set up digital channel direction **/
    for (i = 0; i < DIGCNT; ++i) digital[i] = 0;
    CCALL( "S" , analog, digital );

}



/*-----------------------------------------------------------------
	display  -  Displays the data from analog card.  User aborts
		    with a Ctrl-C.
*/

void    display()
{
	for(;;) 
	{
	   /** Press Ctrl-C to exit the program **/
	   if (kbhit()) (getch());   

       /** Calibrate and read analog inputs **/
       CCALL( "cm" , analog, digital );         /* Call A-D driver */

       /** Read digital inputs **/
       CCALL( "I" , analog, digital );         /* Call A-D driver */

       /** Output channel data **/
	   system("cls");
       printf("\nPress ctrl-C to return to dos.\n\n\n");
       printf("Channel     Analog Input     Digital input\n");
       printf("-------     ------------     -------------\n");
       for (i = 0; i < 8; ++i)
	      printf("%4d %20e %9u\n",
		      i + 1, (double) analog[i],(unsigned int) digital[i]);
	}
}

/*---------------------------------------------------------------- 
			Main program     
*/

void    main()
{
    initial();                          /* Initialize A-D driver */
    display();                          /* Display driver output */
}

