/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                         */
/*      Analog - Digital Demonstration Program for C language              */
/*      Copyright (c) 1986,1988,1989 STI                                   */
/*                                                                         */
/* ----------------------------------------------------------------------- */
/* NOTE:                                                                   */
/*      Small Memory Model:                                                */
/*            Change CCALL() to SH_CALL() and link with MS_CALL.OBJ        */
/*      Large/Huge Memory Model:                                           */
/*            Use CCALL() and link with MS_CALL.OBJ                        */
/*                                                                         */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifdef lattice
typedef int void;
#endif

#include "stdio.h"

#define  ANACNT  8                       /* Number of analog channels */
#define  DIGCNT  8                       /* Number of digital channels */

unsigned digital[DIGCNT];                /* Digital channel array */
unsigned analog[ANACNT];                 /* Analog channel array */
int      i,j;                     



/*-------------------------------------------------------------
	initial()  -  Initialize and calibrate A-D board.  Check 
	              and report any errors to user.
*/

void    initial()
{
    /** Read CALIB.DAT file and calibrate the analog inputs **/

    CCALL( "Fn" , analog, digital );         /* Call A-D driver */

    if (digital[1] == 0 && digital[2] == 0)  {
        printf("\nDriver, ADRIVE.COM, not installed; ");
        printf(" or analog card not installed.\n");
        exit(1);
    }
    if (digital[1] == 0 && digital[2] != 0)  {
        printf("\nNo analog card selected. BRD SEL switch set to 0.\n");
        exit(1);
    }
    if (digital[1] != 0 && digital[4] == 0)  {
        printf("\nCALIB.DAT file not correct or FIND.EXE was not run.\n");
        exit(1);
    }
    if (digital[1] > digital[4])  {
        printf("\nCalibration numbers are not correct.\n");
        exit(1);
    }
    if (digital[1] > ANACNT || digital[2] > DIGCNT)  {
        printf("\nToo many channels installed.\n");
        printf("Change size of ANACNT and/or DIGCNT in heading\n");
        exit(1);
    }

    /** Set up digital channel direction **/
    for (i = 0; i < DIGCNT; ++i) digital[i] = 0;
    CCALL( "S" , analog, digital );   

}



/*---------------------------------------------------------
	display()  -  Displays the data from analog card.  User
                  aborts with a Ctrl-C.
*/

void    display()
{
    while(1)
    {
    /** Get analog output value from user **/
    printf("\n\nPress ctrl-C to return to dos.");
    printf("\nChannel to output analog data: ");
    scanf("%d", &i);
    if (i < 1 || i > ANACNT) i = 1;
    printf("\nData to output (0 to 4095): ");
    scanf("%d", &j);
    if (j < 0 || j > 4095) j = 0;
    CCALL("w", &j, &i);

    /** Read digital inputs **/
    CCALL( "I" , analog, digital );    

    /** Output channel data **/
    printf("Channel     Digital input\n");
    printf("-------     -------------\n");
    for (i = 0; i < 8; ++i)
        printf("%4d %15u\n",
               i + 1, digital[i]);
    }
}

/* * * * * * * * * * * * * * * * * * * */
/*  Main program                       */
/* * * * * * * * * * * * * * * * * * * */

void    main()
{
    initial();                          /* Initialize A-D driver */
    display();                          /* Display driver output */
}

