/************************************************************************
*                                                                       *
*       Counter/Timer Demonstration Program for C language              *
*       Copyright (c) 1993 Strawberry Tree Incorporated    		        *
*                                                                       *
*		REVS	:	02/18/93 Original Release		                    *
*		INCLUDE	:	ctimer.h	     				                    *
*		COMPILER:	Borland Turbo C++ Version 1.00			            *
*  -------------------------------------------------------------------  *
*		DESCRIPTION :                                                   *
*       The Pulsein program is an example program to illustrate the 	*
*		use of Counter/Timer commands in the driver to count incoming	*
*		pulses on the external input using a digital I/O.        	    *
*                                                                       *
*		Refer to the hardware manual for details on the four external   *
*       I/O numbers associated with each Counter/Timer: Input, Output,  *
*       Trigger, and Gate.                      	                    *
*									                                    *
************************************************************************/
#include <stdio.h>
#include "ctimer.h"

unsigned short	Digital[32];
float         	Analog[32];

void SetupCount();
void ReadCount();
void ResetCount();

void
main()
{
	clrscr();
	printf("Put Clean TTL Pulses into the digital I/O line\n");
	printf("that associates with Counter/Timer Input.\n\n");
	printf("Press R to RESTART count.\n");
	printf("Press S to STOP count where it is.\n");
	printf("Press C to CONTINUE count from where you stopped.\n");
	printf("Type CTRL-C to EXIT.\n\n");

	SetupCount();			/* Set mode to count start from initial value */
	for(;;)					/* Forever loop */
	{
		if(kbhit())			/* If you asked for something. */
		{
			switch(getch())
			{
				case 'r':	/* Press r or R to reset count */
				case 'R':
					ResetCount();
					break;
				case 'c':
				case 'C':   /* Press c or C to continue count */
                    /* Set mode to countinue the count */
					Digital[0] = SOFTWARE_TRIG | SOFTWARE_GATE | 
					             EXTERNAL_INPUT | CONTINUOUS;
					Analog[0] = 1.00;
					SH_CALL( "x" , Analog, Digital );
					SetupCount();
					break;
				case 's':
				case 'S':   /* Press s or S to stop count */
					Digital[0] = 0x00;
					Analog[0] = 1.00;
					SH_CALL( "x" , Analog, Digital );
					break;
				default:
					break;
			}
		}
        ReadCount();
	}
}


void
SetupCount()
{
/* Set count value in Counter/Timer 1 */
	Digital[0] = 3000;
	Analog[0] = 1.0;
	SH_CALL( "u" , Analog, Digital );

/* Set the mode for Counter/Timer 1 to count pulses coming 
   into the external input */
	Digital[0] = SOFTWARE_TRIG | SOFTWARE_GATE | IO_ENABLE |
	             EXTERNAL_INPUT | CONTINUOUS;
	Analog[0] = 1.0;
	SH_CALL( "x" , Analog, Digital );
}


void
ReadCount()
{
/* Read the current value of Counter/Timer 1 */
	Analog[0] = 1.0;
	SH_CALL( "v" , Analog, Digital );
    printf("\rPulses counted so far %u        ",Digital[0]);
}


void
ResetCount()
{
/* Reload the intial value*/
	Digital[0] = 3000;
	Analog[0] = 1.0;
	SH_CALL( "u" , Analog, Digital );

/* Set the mode for Counter/Timer to retrigger */
	Digital[0] = SOFTWARE_TRIG | RETRIG_EN;
	Analog[0] = 1.0;
	SH_CALL( "x" , Analog, Digital );

/* Set the mode for Counter/Timer 1 to count pulses coming 
   into the external input */
	Digital[0] = SOFTWARE_TRIG | SOFTWARE_GATE | IO_ENABLE |
				 EXTERNAL_INPUT | CONTINUOUS;
	Analog[0] = 1.0;
	SH_CALL( "x" , Analog, Digital );

}
