/*  CXL - Copyright (c) 1987-1989 by Mike Smedley - All Rights Reserved  */
/*  WDRAG.C - smoothly drag a window 1 position in given direction  */

#include <stdlib.h>
#include <string.h>
#include "cxlvid.h"
#include "cxlwin.h"

int wdrag(int direction)
{
    register int *src,*dest;
    int *win_image,*wp,*p;
    int nsrow,nscol,nerow,necol,shad_attr=-1;
    int srow,scol,erow,ecol,fill_row,fill_col,i;
    int vert_movement,horz_movement;
    int chars_per_line,lines_per_win;

    /* check for active window */
    if(!_winfo.total) return(_winfo.errno=W_NOACTIVE);

    /* get window coordinates */
    srow = _winfo.active->srow;
    scol = _winfo.active->scol;
    erow = _winfo.active->erow;
    ecol = _winfo.active->ecol;

    /* calculate lines-per-window and characters-per-line */
    lines_per_win  = erow - srow + 1;
    chars_per_line = ecol - scol + 1;

    /* determine facts about direction of move */
    vert_movement=horz_movement=0;
    switch(direction) {
        case D_DOWN:
            vert_movement=1;
            lines_per_win--;
            break;
        case D_UP:
            vert_movement=-1;
            lines_per_win--;
            break;
        case D_LEFT:
            horz_movement=-1;
            chars_per_line--;
            break;
        default:    /* D_RIGHT */
            horz_movement=1;
            chars_per_line--;
    }

    /* calculate new window coordinates */
    nsrow = srow + vert_movement;
    nscol = scol + horz_movement;
    nerow = erow + vert_movement;
    necol = ecol + horz_movement;

    /* if window has shadow, close it before the move */
    if(_winfo.active->wsbuf!=NULL) {
        shad_attr=_winfo.active->wsattr;
        wshadoff();
    }

    /* save the current window image */
    if((win_image=wsave(srow,scol,erow,ecol))==NULL)
        return(_winfo.errno=W_ALLOCERR);

    /* save the area where the window will relocate to */
    if((wp=wsave(nsrow,nscol,nerow,necol))==NULL) {
        free(win_image);
        return(_winfo.errno=W_ALLOCERR);
    }

    /* change coordinates in saved window's buffer */
    /* and restore window to new coordinates       */
    *(win_image)   = nsrow;
    *(win_image+1) = nscol;
    *(win_image+2) = nerow;
    *(win_image+3) = necol;
    wrestore(win_image);

    /* start buffer positions past coordinates */
    src  = _winfo.active->wbuf + 4;
    dest = wp + 4;

    if(direction==D_DOWN) src  += chars_per_line;
    if(direction==D_UP)   dest += chars_per_line;

    /* do the transfer of buffer contents */
    for (i=0; i < lines_per_win; i++) {

        if(direction==D_LEFT)  dest++;
        if(direction==D_RIGHT) src++;

        /* move 1 line */
        memmove(dest,src,chars_per_line*sizeof(int));
        src  += chars_per_line;
        dest += chars_per_line;

        if(direction==D_LEFT)  src++;
        if(direction==D_RIGHT) dest++;
    }

    /* erase the trail that was left-over */
    p = _winfo.active->wbuf + 4;
    if(vert_movement) {
        if(direction==D_DOWN)
            fill_row=srow;
        else {
            p += (lines_per_win * chars_per_line);
            fill_row=erow;
        }
        for(i=scol;i<=ecol;i++,p++) printc(fill_row,i,(*p)>>8,(*p)&0x00ff);
    }
    else {
        if(direction==D_RIGHT)
            fill_col=scol;
        else {
            p += chars_per_line;
            fill_col=ecol;
        }
        for(i=srow;i<=erow;i++,p+=chars_per_line+1)
            printc(i,fill_col,(*p)>>8,(*p)&0x00ff);
    }

    /* free old window buffer */
    free(_winfo.active->wbuf);

    /* update window record */
    _winfo.active->wbuf   = wp;
    _winfo.active->row    = _winfo.active->row    - _winfo.active->srow+ nsrow;
    _winfo.active->column = _winfo.active->column - _winfo.active->scol+ nscol;
    _winfo.active->srow   = nsrow;
    _winfo.active->scol   = nscol;
    _winfo.active->erow   = nerow;
    _winfo.active->ecol   = necol;

    /* if window has shadow, redraw it */
    if(shad_attr!=-1) wshadow(shad_attr);

    /* reset cursor position */
    gotoxy_(_winfo.active->row,_winfo.active->column);

    /* return normally */
    return(_winfo.errno=W_NOERROR);
}
