/*  CXL - Copyright (c) 1987-1990 by Mike Smedley - All Rights Reserved  */
/* WHELP.C  -  whelpdef()  defines the help key, file, and window colors
               whelpwin()  defines the help window boundaries
               whelpcat()  sets the help category
               whelppcat() sets the help category, but pushes current one first
               whelpopc()  pops the last saved help category into the void
               whelpclr()  clears the help buffer */

#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cxldef.h"
#include "cxlkey.h"
#include "cxlmou.h"
#include "cxlstr.h"
#include "cxlvid.h"
#include "cxlwin.h"

#if defined(__TURBOC__)                         /* Turbo C */
    #include <dir.h>
#else                                           /* all others */
    #include <direct.h>
#endif

#if defined(__ZTC__)                            /* Zortech C/C++ */
    #define stricmp(a,b) strcmpl(a,b)
#endif

#define BUFSIZE     80
#define BASETAGID   12345   /* base tag ID to use for cross reference items */
#define MAXXREF     30      /* max cross references per help screen */

/* function prototypes */
static void disp_cat(void);
static void esc_esc(void);
static void esc_pgdn(void);
static void esc_pgup(void);
static int  find_cat_name(char *cat);
static int  find_cat_number(int cat);
static int  find_page(long startpos,int pageofs);
static void help_handler(void);
static void not_found(char *cat);

/* static variables (local to this file) */
static char *_Near _Cdecl catarray[MAXXREF];
static int   _Near _Cdecl arraycnt=0;
static char  _Near _Cdecl buf[BUFSIZE+1];
static FILE *_Near _Cdecl fp;
static char  _Near _Cdecl line1[]="";
static char  _Near _Cdecl line2[]="PgUp";
static char  _Near _Cdecl line3[]="PgDn";

static struct _help_t _Near _Cdecl whelp=
{
    { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 },
    NULL,-1,0,0,0,0,0,3,8,21,71,0,YES,NULL
};

/* help index file record definition */
struct recd_t {
    int help;
    char category[30];
    long offset;
};

static struct recd_t _Near _Cdecl recd;

/*---------------------------------------------------------------------------*/

int whelpdef(char *file,unsigned key,int winattr,int textattr,int selattr,
             int barattr,void (*open)(void))
{
    /* is help disengagement requested?  If so, un-define the help key. */
    if(file==NULL) {
        if(_winfo.helptr==NULL)
            return(_winfo.errno=W_NOHLPDEF);
        else {
            _winfo.helptr=NULL;
            whelpclr();
            setonkey(whelp.key,NULL,0);
        }
    }
    else {

        /* attach help key to help handler */
        if(setonkey(key,help_handler,0)) return(_winfo.errno=W_ALLOCERR);

        /* convert file name to upper case */
        strupr(file);

        /* point global help info pointer to internal struct */
        _winfo.helptr=&whelp;
    }

    /* add information to window help information record */
    whelp.file     = file;
    whelp.key      = key;
    whelp.winattr  = mapattr(winattr);
    whelp.textattr = mapattr(textattr);
    whelp.selattr  = mapattr(selattr);
    whelp.barattr  = mapattr(barattr);
    whelp.open     = open;

    /* return with no error */
    return(_winfo.errno=W_NOERROR);
}

/*---------------------------------------------------------------------------*/

int whelpwin(int srow,int scol,int erow,int ecol,int btype,int title)
{
    /* make sure help has been defined */
    if(_winfo.helptr==NULL) return(_winfo.errno=W_NOHLPDEF);

    /* load given coordinates into local static structure */
    whelp.srow =srow;
    whelp.scol =scol;
    whelp.erow =erow;
    whelp.ecol =ecol;
    whelp.btype=btype;
    whelp.title=title;

    /* return normally */
    return(_winfo.errno=W_NOERROR);
}

/*---------------------------------------------------------------------------*/

int whelpcat(int cat)
{
    /* make sure help record has been defined */
    if(_winfo.helptr==NULL) return(_winfo.errno=W_NOHLPDEF);

    /* determine if help is to be global or window */
    if(_winfo.active!=NULL) _winfo.active->help=cat;
    _winfo.help=cat;

    /* return normally */
    return(_winfo.errno=W_NOERROR);
}

/*---------------------------------------------------------------------------*/

int whelppcat(int cat)
{
    /* check for existance of help, and push */
    /* current help category onto help stack */
    if(whelpush()) return(_winfo.errno);

    /* set help category */
    return(whelpcat(cat));
}

/*---------------------------------------------------------------------------*/

int whelpopc(void)
{
    /* make sure help has been defined */
    if(_winfo.helptr==NULL) return(_winfo.errno=W_NOHLPDEF);

    /* check for stack underflow */
    if(whelp.helpptr==-1) return(_winfo.errno=W_HLPSTKUN);

    /* decrement stack pointer */
    whelp.helpptr--;

    /* return normally */
    return(_winfo.errno=W_NOERROR);
}

/*---------------------------------------------------------------------------*/

int whelpclr(void)
{
    /* make sure help has been defined */
    if(_winfo.helptr==NULL) return(_winfo.errno=W_NOHLPDEF);

    /* clear help category and reset help stack pointer */
    _winfo.help=0;
    whelp.helpptr=-1;

    /* return normally */
    return(_winfo.errno=W_NOERROR);
}

/*---------------------------------------------------------------------------*/

static void disp_cat(void)
{
    int i,newpage,page,wrow,wcol,end,menuopen,itemopen;
    long startpos;
    struct _menu_t *curr;
    char *p,*q;

    /* initialize variables */
    page=wrow=wcol=end=menuopen=itemopen=0;

    /* save current info */
    startpos = ftell(fp);
    curr     = _winfo.cmenu;

    /* set text attribute */
    wtextattr(whelp.textattr);

    for(;;) {

        /* read next line from help file into buffer */
        fgets(buf,BUFSIZE,fp);

        /* if end-of-file or "*E" was found, assume end-of-category */
        end=( !strnicmp(buf,"*E",2) || feof(fp) )?YES:NO;

        /* if end-of-category or new-page specified */
        if( !strnicmp(buf,"*P",2) || end ) {
loop:
            /* update bottom border of window with PgUp/PgDn info */
            if(page&&!end)       p="PgUp/PgDn";
            else if(!page&&end)  p=line1;
            else if(page&&end)   p=line2;
            else                 p=line3;
	    wmessage(p,BT_BORD,40,whelp.winattr);

            /* if one or more cross reference menu items were specified */
            if(arraycnt) {

                /* depending on whether there are previous  */
                /* or next pages, define the PgUp/PgDn keys */
                if(page) setonkey(0x4900,esc_pgup,0);
                if(!end) setonkey(0x5100,esc_pgdn,0);

                /* define the Esc key */
                setonkey(0x011b,esc_esc,0);

                /* end the menu and process it */
                wmenuend(BASETAGID,M_OMNI|M_NOQS,0,0,whelp.selattr
                  ,whelp.selattr,0,whelp.barattr);
                i=wmenuget();

                /* free the keys that were defined */
                if(page) setonkey(0x4900,NULL,0);
                if(!end) setonkey(0x5100,NULL,0);
                setonkey(0x011b,NULL,0);

                /* turn off the menuopen flag and restore menu info */
                menuopen=NO;
                _winfo.cmenu=curr;

                /* if Esc, PgDn, or PgUp was NOT pressed */
                if(i!=-1) {

                    /* convert tagid to array subscript */
                    i-=BASETAGID;

                    /* try to find selected category, if found set */
                    /* file pointer to it, otherwise reset file    */
                    /* pointer back to previous help category      */
                    if(find_cat_name(catarray[i]))
                        startpos=ftell(fp);
                    else
                        fseek(fp,startpos,SEEK_SET);

                    /* clear help window and set     */
                    /* position to upper left corner */
                    wclear();
                    wrow=page=0;

                    /* free menu item strings */
                    for(i=0;i<arraycnt;i++) free(catarray[i]);
                    arraycnt=0;
                    continue;
                }

                /* free menu item strings */
                for(i=0;i<arraycnt;i++) free(catarray[i]);
                arraycnt=0;
            }

            /* get keypress and test it */
            switch(getxch()) {

                case 0x4900:    /* PgUp */

                    /* find previous page, clear help window, */
                    /* and set position to upper left corner  */
                    newpage=find_page(startpos,page-1);
                    if(newpage==page) break;
                    page=newpage;
                    wrow=0;
                    wclear();
                    continue;

                case 0x5100:    /* PgDn */

                    /* find next page, clear help window, and */
                    /* set position to upper left corner      */
                    newpage=find_page(startpos,page+1);
                    if(newpage==page) break;
                    page=newpage;
                    wrow=0;
                    wclear();
                    continue;

                case 0x011b:    /* Esc */

                    /* clear any PgUp or PgDn keystrokes that */
                    /* may be lurking in CXL's keystroke      */
                    /* buffer, and return to caller.          */
                    kbclear();
                    return;
            }

            /* a goto statement is used here so the continue */
            /* statement above falls out to the right loop   */
            goto loop;
        }

        /* search for cross reference categories on current line. */
        /* if any are found, then define them as a menu item      */
        q=buf;
        do {

            /* search for 1st carat. if found, test for double carat. */
            /* if double carat was found display literal carat. next, */
            /* search for 2nd carat.  if not found, then display the  */
            /* line as text.  otherwise, if found, define everything  */
            /* in between the 2 carats as a menu item.                */
            p=strchr(q,'^');
            if(p!=NULL) {
                if(*(p+1)=='^')  {
                    *(p+1)='\0';
                    wputs(q);
                    wcol=wcol+strlen(q);
                    q=p+2;
                }
                else {
                    *p='\0';
                    itemopen^=1;
                    wputs(q);
                    if(!itemopen) {
                        if((catarray[arraycnt]=malloc(strlen(q)+1))!=NULL) {
                            strcpy(catarray[arraycnt],q);
                            if(!menuopen) { wmenubegc(); menuopen=YES; }
                            wmenuitem(wrow,wcol,catarray[arraycnt],
                              *catarray[arraycnt],BASETAGID+arraycnt,0,NULL,0
                              ,0);
                            arraycnt++;
                        }
                    }
                    wcol=wcol+strlen(q);
                    q=p+1;
                }
            }
        } while(p!=NULL);

        /* display text line and increment current window row */
        wputs(q);
        wrow++;
        wcol=0;
    }
}

/*---------------------------------------------------------------------------*/

static void esc_esc(void)
{
    kbput(0x011b);
    kbput(0x011b);
    setonkey(0x011b,NULL,0);
}

/*---------------------------------------------------------------------------*/

static void esc_pgdn(void)
{
    kbput(0x011b);
    kbput(0x5100);
    setonkey(0x5100,NULL,0);
}

/*---------------------------------------------------------------------------*/

static void esc_pgup(void)
{
    kbput(0x011b);
    kbput(0x4900);
    setonkey(0x4900,NULL,0);
}

/*---------------------------------------------------------------------------*/

static int find_cat_name(char *cat)
{
    int found=NO;

    /* Reset file pointer. */
    rewind(fp);

    /* Check for "*I" marker. */
    fgets(buf,BUFSIZE,fp);
    if(!strnicmp(buf,"*I",2)) {

        /* Search index for help category entry.  If found, */
        /* then advance file pointer to specified position. */
        for(;;) {
            fread(&recd,sizeof(struct recd_t),1,fp);
            if(feof(fp)||recd.offset==-1L) break;
            if(!stricmp(recd.category,cat)) {
                fseek(fp,recd.offset,SEEK_SET);
                found=YES;
                break;
            }
        }
    }

    /* If help category was not found, display an error message. */
    if(!found) not_found(cat);

    return(found);
}

/*---------------------------------------------------------------------------*/

static int find_cat_number(int cat)
{
    int found=NO;

    /* Reset file pointer. */
    rewind(fp);

    /* Check for "*I" marker. */
    fgets(buf,BUFSIZE,fp);
    if(!strnicmp(buf,"*I",2)) {

        /* Search index for help category entry.  If found, */
        /* then advance file pointer to specified position. */
        for(;;) {
            fread(&recd,sizeof(struct recd_t),1,fp);
            if(feof(fp)||recd.offset==-1L) break;
            if(recd.help==cat) {
                fseek(fp,recd.offset,SEEK_SET);
                found=YES;
                break;
            }
        }
    }

    /* If help category was not found, display an error message. */
    if(!found) not_found(itoa(cat,buf,10));

    return(found);
}

/*---------------------------------------------------------------------------*/

static int find_page(long startpos,int pageofs)
{
    long lastpagepos,currpos;
    int currpage=0;

    lastpagepos=currpos=startpos;
    fseek(fp,startpos,SEEK_SET);
    while(currpage<pageofs) {
        fgets(buf,BUFSIZE,fp);
        currpos=ftell(fp);
        if( !strnicmp(buf,"*E",2) || feof(fp)) {
            fseek(fp,lastpagepos,SEEK_SET);
            break;
        }
        if(!strnicmp(buf,"*P",2)) {
            lastpagepos=currpos;
            currpage++;
        }
    }
    return(currpage);
}

/*---------------------------------------------------------------------------*/

static void help_handler(void)
{
    register int i;
    int help,cat,linechar,mouse,escchk;
    int found;
    struct _onkey_t *savekb;
    extern char *_box_table[];

    /* save help category */
    help=_winfo.help;

    /* save escape checking value, and allow [Esc] key */
    escchk=_winfo.esc;
    _winfo.esc=TRUE;

    /* temporarily un-define the onkey list to avoid collision */
    savekb=chgonkey(NULL);

    /* hide mouse cursor */
    if(_mouse&MS_CURS) mshidecur();

    /* open help window */
    if(wopen(whelp.srow,whelp.scol,whelp.erow,whelp.ecol,whelp.btype
      ,whelp.winattr,whelp.winattr)) {

        /* display window title if specified */
        if(whelp.title) wtitle("[ Help ]",TCENTER,whelp.winattr);

        /* see if an open function was given.  If so, then call it */
        if(whelp.open!=NULL) (*whelp.open)();

        /* set line characters according to window's box type */
        linechar=_box_table[_winfo.active->btype][6];
        strchg(line1,'',linechar);
        strchg(line2,'',linechar);
        strchg(line3,'',linechar);

        /* disable mouse cursor */
        mouse=_mouse;
        mssupport(MS_KEYS);

        /* hide cursor */
        hidecur();

        /* determine help category to use.  If the current help category */
        /* is empty, check the stack for a help category there           */
        cat=help;
        if(!cat) {
            if(whelp.helpptr>-1) {
                for(i=whelp.helpptr;i>=0;i--) {
                    if(whelp.help[i]) {
                        cat=whelp.help[i];
                        break;
                    }
                }
            }
        }

        /* check for no defined help category */
        if(!cat) {
            wtextattr(whelp.textattr);
            wputs("No help category defined.\nPress a key to continue.");
            waitkey();
        }
        else {

            /* open help file */
            if((fp=fopen(whelp.file,"rb"))==NULL) {
                wtextattr(whelp.textattr);
                wputs("Help file not found:  ");
                wputs(whelp.file);
                wputs("\nPress a key to continue.");
                waitkey();
            }
            else {

                /* find help category in help file */
                found=find_cat_number(cat);

                /* read and display help category text */
                if(found) disp_cat();

                /* close help file */
                fclose(fp);
            }
        }

        /* close help window */
        wclose();
    }

    /* reveal hidden cursor */
    showcur();

    /* reset escape checking value */
    _winfo.esc=escchk;

    /* reset mouse info */
    if((_mouse=mouse)&MS_CURS) msshowcur();

    /* clear any onkeys defined in "open" function and relink the onkey list */
    freonkey();
    chgonkey(savekb);

    /* restore help category */
    _winfo.help=help;
}

/*---------------------------------------------------------------------------*/

static void not_found(char *cat)
{
    wtextattr(whelp.textattr);
    wputs("Help category not found:  ");
    wputs(cat);
    wputs("\nPress a key to continue.");
    waitkey();
}
