/*  CXL - Copyright (c) 1987-1989 by Mike Smedley - All Rights Reserved  */
/*  WHLINE.C    - whline() draws a horizontal text line in active window
                  wvline() draws a vertical text line in active window    */

#include <dos.h>
#include "cxldef.h"
#include "cxlvid.h"
#include "cxlwin.h"

#define HORZ  0
#define VERT  1

#define ULC   bt[0]    /* upper left corner         */
#define UHL   bt[1]    /* upper horizontal line     */
#define URC   bt[2]    /* upper right corner        */
#define LVL   bt[3]    /* left vertical line        */
#define RVL   bt[4]    /* right vertical line       */
#define LLC   bt[5]    /* lower left corner         */
#define LHL   bt[6]    /* lower horizontal line     */
#define LRC   bt[7]    /* lower right corner        */
#define MJ    bt[8]    /* middle junction           */
#define LVJ   bt[9]    /* left vertical junction    */
#define RVJ   bt[10]   /* right vertical junction   */
#define UHJ   bt[11]   /* upper horizontal junction */
#define LHJ   bt[12]   /* lower horizontal junction */

extern char *_box_table[];

/* function prototypes */
static int disp_char(int wrow,int wcol,int attr,int btype,int ch,int direc);
static int isupvert(int btype,int c);
static int isdownvert(int btype,int c);
static int islefthorz(int btype,int c);
static int isrighthorz(int btype,int c);
static int read_char(int wrow,int wcol);

/*---------------------------------------------------------------------------*/

int whline(int wsrow,int wscol,int count,int btype,int attr)
{
    register char *bt;
    register int col;
    int row,up,down;
    char ch;

    /* check for active window */
    if(!_winfo.total) return(_winfo.errno=W_NOACTIVE);

    /* check for valid box type */
    if(btype<0||btype>5) return(_winfo.errno=W_INVBTYPE);
    row=wsrow; col=wscol;

    /* abbreviate pointer */
    bt=_box_table[btype];

    if(count) {

        /* see if a left junction or corner is needed */
        up=isupvert(btype,read_char(row-1,col));
        down=isdownvert(btype,read_char(row+1,col));
        if(up&&down)
            ch=LVJ;
        else if(up)
            ch=LLC;
        else if(down)
            ch=ULC;
        else
            ch=UHL;

        /* display leftmost character */
        if(disp_char(row,col,attr,btype,ch,HORZ)) return(_winfo.errno);
        col++;
        count--;
    }

    /* do while not last character */
    while(count>1) {

        /* see if a middle junction is needed */
        up=isupvert(btype,read_char(row-1,col));
        down=isdownvert(btype,read_char(row+1,col));
        if(up&&down)
            ch=MJ;
        else if(up)
            ch=LHJ;
        else if(down)
            ch=UHJ;
        else
            ch=UHL;

        /* display middle character */
        if(disp_char(row,col,attr,btype,ch,HORZ)) return(_winfo.errno);
        col++;
        count--;
    }

    if(count) {

        /* see if a right junction or corner is needed */
        up=isupvert(btype,read_char(row-1,col));
        down=isdownvert(btype,read_char(row+1,col));
        if(up&&down)
            ch=RVJ;
        else if(up)
            ch=LRC;
        else if(down)
            ch=URC;
        else
            ch=UHL;

        /* display rightmost character */
        if(disp_char(row,col,attr,btype,ch,HORZ)) return(_winfo.errno);
    }

    /* return normally */
    return(_winfo.errno=W_NOERROR);
}

/*---------------------------------------------------------------------------*/

int wvline(int wsrow,int wscol,int count,int btype,int attr)
{
    register char *bt;
    register int col;
    int row,left,right;
    char ch;

    /* check for active window */
    if(!_winfo.total) return(_winfo.errno=W_NOACTIVE);

    /* check for valid box type */
    if(btype<0||btype>5) return(_winfo.errno=W_INVBTYPE);
    row=wsrow; col=wscol;

    /* abbreviate pointer */
    bt=_box_table[btype];

    if(count) {

        /* see if a top junction or corner is needed */
        left=islefthorz(btype,read_char(row,col-1));
        right=isrighthorz(btype,read_char(row,col+1));
        if(left&&right)
            ch=UHJ;
        else if(left)
            ch=URC;
        else if(right)
            ch=ULC;
        else
            ch=LVL;

        /* display uppermost character */
        if(disp_char(row,col,attr,btype,ch,VERT)) return(_winfo.errno);
        row++;
        count--;
    }

    /* do while not last character */
    while(count>1) {
        left=islefthorz(btype,read_char(row,col-1));
        right=isrighthorz(btype,read_char(row,col+1));
        if(left&&right)
            ch=MJ;
        else if(left)
            ch=RVJ;
        else if(right)
            ch=LVJ;
        else
            ch=LVL;

        /* display middle character */
        if(disp_char(row,col,attr,btype,ch,VERT)) return(_winfo.errno);
        row++;
        count--;
    }

    if(count) {

        /* see if a bottom junction or corner is needed */
        left=islefthorz(btype,read_char(row,col-1));
        right=isrighthorz(btype,read_char(row,col+1));
        if(left&&right)
            ch=LHJ;
        else if(left)
            ch=LRC;
        else if(right)
            ch=LLC;
        else
            ch=LVL;

        /* display bottommost character */
        if(disp_char(row,col,attr,btype,ch,VERT)) return(_winfo.errno);
    }

    /* return normally */
    return(_winfo.errno=W_NOERROR);
}

/*---------------------------------------------------------------------------*/

static int disp_char(int wrow,int wcol,int attr,int btype,int ch,int direc)
{
    register char *bt;
    register int col;
    int row;

    /* see if next to a border, if so, connect to it */
    if(_winfo.active->border) {

        /* abbreviate pointer */
        bt=_box_table[btype];

        /* calculate effective row and column */
        row=_winfo.active->srow+_winfo.active->border+wrow;
        col=_winfo.active->scol+_winfo.active->border+wcol;

        /* see if this is a horizontal or vertical line */
        if(direc==HORZ) {

            /* make sure that the box type characters match */
            if(LVL==_box_table[_winfo.active->btype][3]) {

                /* check left border */
                if(col==(_winfo.active->scol+1)) {
                    printc(row,_winfo.active->scol,attr,LVJ);
                    ch=UHL;
                }

                /* check right border */
                if(col==(_winfo.active->ecol-1)) {
                    printc(row,_winfo.active->ecol,attr,RVJ);
                    ch=UHL;
                }
            }
        }
        else {

            /* make sure that the box type characters match */
            if(UHL==_box_table[_winfo.active->btype][1]) {

                /* check top border */
                if(row==(_winfo.active->srow+1)) {
                    printc(_winfo.active->srow,col,attr,UHJ);
                    ch=LVL;
                }

                /* check bottom border */
                if(row==(_winfo.active->erow-1)) {
                    printc(_winfo.active->erow,col,attr,LHJ);
                    ch=LVL;
                }
            }
        }
    }

    /* display character */
    if(wprintc(wrow,wcol,attr,ch)) return(_winfo.errno);

    /* return normally */
    return(0);
}

/*---------------------------------------------------------------------------*/

static int isupvert(int btype,int c)
{
    register char *bt;
    char ch;

    bt=_box_table[btype];
    ch=(char)c;
    return((ch==LVL||ch==UHJ||ch==ULC||ch==URC||ch==LVJ||ch==RVJ||ch==MJ)
        ?YES:NO);
}

/*---------------------------------------------------------------------------*/

static int isdownvert(int btype,int c)
{
    register char *bt;
    char ch;

    bt=_box_table[btype];
    ch=(char)c;
    return((ch==LVL||ch==LHJ||ch==LLC||ch==LRC||ch==LVJ||ch==RVJ||ch==MJ)
           ?YES:NO);
}

/*---------------------------------------------------------------------------*/

static int islefthorz(int btype,int c)
{
    register char *bt;
    char ch;

    bt=_box_table[btype];
    ch=(char)c;
    return((ch==UHL||ch==LVJ||ch==LLC||ch==ULC||ch==UHJ||ch==LHJ||ch==MJ)
           ?YES:NO);
}

/*---------------------------------------------------------------------------*/

static int isrighthorz(int btype,int c)
{
    register char *bt;
    char ch;

    bt=_box_table[btype];
    ch=(char)c;
    return((ch==UHL||ch==RVJ||ch==LRC||ch==URC||ch==UHJ||ch==LHJ||ch==MJ)
           ?YES:NO);
}

/*---------------------------------------------------------------------------*/

static int read_char(int wrow,int wcol)
{
    int ch;
    char far *p;

    if(_vinfo.usebios) {
        wgotoxy(wrow,wcol);
        ch=readchat();
    }
    else {
        /* calculate video RAM address */
        p=MK_FP(_vinfo.videoseg,(((_winfo.active->srow+wrow+_winfo.active->border)*_vinfo.numcols)+(_winfo.active->scol+wcol+_winfo.active->border))*2);
        if(_vinfo.cgasnow)
            ch=cgardbyte(p);
        else
            ch=*p;
    }
    return(ch&0x00ff);
}
