/*  CXL - Copyright (c) 1987-1989 by Mike Smedley - All Rights Reserved  */
/*  WINP.C  - winpbeg()  defines a window input form record
              winpdef()  defines a window input field record
              winpread() reads all defined input fields for current form */

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "cxldef.h"
#include "cxlkey.h"
#include "cxlstr.h"
#include "cxlvid.h"
#include "cxlwin.h"

/* function prototypes */
static void             adv_column(struct _form_t *cform,int col);
static void             after_edit(struct _form_t *cform);
static void             back_space(struct _form_t *cform);
static void             beg_pos(struct _form_t *cform);
static void             call_func(void (*func)(void));
static void             check_format(char *format,int *lenfld,int *lenbuf,
                                     int *decpos);
static void             check_redisp(struct _form_t *cform);
static int              closest_column(struct _form_t *cform,
                                       struct _field_t *field);
static int              cond_update(struct _form_t *cform);
static void             deci_left(struct _form_t *cform,char *start);
static void             deci_right(struct _form_t *cform,char *start);
static void             del_char(struct _form_t *cform);
static void             del_field(struct _form_t *cform,char *start,int high);
static void             del_left_word(struct _form_t *cform);
static void             del_rest(struct _form_t *cform);
static void             del_word(struct _form_t *cform);
static void             disp_char(struct _form_t *cform,int ch,int advance);
static void             disp_field(struct _form_t *cform,char *start,int high);
static void             down_field(struct _form_t *cform);
static void             end_line(struct _form_t *cform);
static void             end_pos(struct _form_t *cform);
static struct _field_t *first_rec(struct _form_t *cform);
static void             free_fields(struct _form_t *cform);
static int              goto_field(struct _form_t *cform,int which);
static void             insert_char(struct _form_t *cform);
static struct _field_t *last_rec(struct _form_t *cform);
static void             next_fchar(struct _form_t *cform);
static void             next_pos(struct _form_t *cform);
static struct _field_t *next_rec(struct _form_t *cform);
static void             next_word(struct _form_t *cform);
static void             prev_fchar(struct _form_t *cform);
static void             prev_pos(struct _form_t *cform);
static struct _field_t *prev_rec(struct _form_t *cform);
static void             prev_word(struct _form_t *cform);
static void             restore_field(struct _form_t *cform,int high);
static void             set_term_key(struct _form_t *cform,unsigned xch);
static void             up_field(struct _form_t *cform);
static int              validate_field(struct _form_t *cform);

/* field movement definitions */
#define FLD_FR  0
#define FLD_LS  1
#define FLD_PR  2
#define FLD_NR  3
#define FLD_UP  4
#define FLD_DN  5

/* array of function pointers to some of the field movement functions */
static struct _field_t *(*funcs[6])(struct _form_t *) =
    { first_rec,last_rec,prev_rec,next_rec };

/*---------------------------------------------------------------------------*/

int winpbeg(int fieldattr,int textattr)
{
    struct _form_t *form;

    /* check for active window */
    if(!_winfo.total) return(_winfo.errno=W_NOACTIVE);

    /* allocate space for new input form record */
    if((form=malloc(sizeof(struct _form_t)))==NULL)
        return(_winfo.errno=W_ALLOCERR);

    /* add new form record to linked list */
    if(_winfo.active->form!=NULL)
        _winfo.active->form->next=form;
    form->prev=_winfo.active->form;
    form->next=NULL;
    _winfo.active->form=form;

    /* adjust attributes if monochrome mapping is on */
    textattr=mapattr(textattr);
    fieldattr=_vinfo.mapattr?revsattr(textattr):fieldattr;

    /* save info in new form record */
    form->field     = NULL;
    form->getkey    = NULL;
    form->termkey   = NULL;
    form->decimal   =
    form->insert    = OFF;
    form->textattr  = textattr;
    form->fieldattr = fieldattr;

    /* return with no error */
    return(_winfo.errno=W_NOERROR);
}

/*---------------------------------------------------------------------------*/

int winpdef(int wrow,int wcol,char *str,char *format,int fconv,int mode,
            int (*validate) (char *),int help)
{
    int lenbuf,lenfld,decpos;
    char *buf;
    struct _field_t *field;

    /* check and see if winpbeg() was called */
    if(_winfo.active->form==NULL) return(_winfo.errno=W_NOFRMBEG);

    /* try to set cursor coordinates - check for active window */
    if(wgotoxy(wrow,wcol)) return(_winfo.errno);

    /* get length of input field and syntax check format string */
    check_format(format,&lenfld,&lenbuf,&decpos);
    if(!lenbuf) return(_winfo.errno=W_INVFORMT);

    /* if field is in update mode, adjust input    */
    /* string size to match "size" of input buffer */
    if(mode) strsetsz(str,lenbuf);

    /* allocate space for buffer */
    if((buf=malloc(lenbuf+1))==NULL) return(_winfo.errno=W_ALLOCERR);

    /* allocate space for new input field record */
    if((field=malloc(sizeof(struct _field_t)))==NULL)
        return(_winfo.errno=W_ALLOCERR);

    /* add new record to linked list */
    if(_winfo.active->form->field!=NULL)
        _winfo.active->form->field->next=field;
    field->prev=_winfo.active->form->field;
    field->next=NULL;
    _winfo.active->form->field=field;

    /* if not updating, then initialize receiving string to spaces */
    if(!mode) {
        memset(str,' ',lenbuf);
        *(str+lenbuf)='\0';
    }

    /* initialize temp buffer to string contents */
    strcpy(buf,str);

    /* save input field record */
    field->wrow     = wrow;
    field->wcol     = wcol;
    field->str      = str;
    field->buf      = buf;
    field->format   = format;
    field->fconv    = fconv;
    field->mode     = mode;
    field->validate = validate;
    field->lenfld   = lenfld;
    field->lenfmt   = strlen(format);
    field->lenbuf   = lenbuf;
    field->decpos   = decpos;
    field->help     = help;
    field->redisp   = OFF;
    field->before   =
    field->after    = NULL;

    /* set current field pointer to new field record */
    _winfo.active->form->cfield=field;

    /* if numeric field, format accordingly */
    if(fconv=='9') deci_right(_winfo.active->form,buf+lenbuf);

    /* display input field */
    disp_field(_winfo.active->form,NULL,0);

    /* return with no error */
    return(_winfo.errno=W_NOERROR);
}

/*---------------------------------------------------------------------------*/

int winpread(void)
{
    struct _form_t  *cform;
    struct _field_t *cfield,*temp;
    char *p;
    int valid,ch,edit,done;
    unsigned int xch;

    /* check for active window */
    if(!_winfo.total) return(_winfo.errno=W_NOACTIVE);

    /* abbreviate pointer to current form */
    cform=_winfo.active->form;

    /* check and see if winpbeg() was called */
    if(cform==NULL) return(_winfo.errno=W_NOFRMBEG);

    /* check for defined input fields */
    if(cform->field==NULL) return(_winfo.errno=W_NOINPDEF);

    /* go to 1st position of first field */
    cform->cfield=first_rec(cform);
    beg_pos(cform);

    /* set help category to 1st field's help category */
    _winfo.help=cform->cfield->help;

    /* call 1st field's before function */
    call_func(cform->cfield->before);

    /* if conditional update, then start cursor end of line */
    edit=cond_update(cform);

    /* initialize Insert mode to OFF */
    cform->insert=OFF;
    smcursor();

    /* initially display field */
    disp_field(cform,NULL,1);

    for(;;) {

        /* set cursor coordinates */
        wgotoxy(cform->cwrow,cform->cwcol);

        /* If alternate getkey function was provided, call it. */
        /* Otherwise call the standard CXL getxch() function   */
        /* If alternate getkey function returned "done" flag,  */
        /* then set termination key variable and exit function */
        if(cform->getkey!=NULL) {
            done=NO;
            xch=(*(cform->getkey))(&done);
            if(done) {
                set_term_key(cform,xch);
                goto form_done;
            }
        }
        else
            xch=getxch();

        /* test keypress */
        switch(xch) {

            case 0x011b:    /* Esc */

                /* if Esc checking is on, restore all fields to their */
                /* original values, free input records, and exit      */
                if(_winfo.esc) {
                    after_edit(cform);
                    for(cform->cfield=cform->field;cform->cfield!=NULL;
                      cform->cfield=cform->cfield->prev)
                        restore_field(cform,0);
                    free_fields(cform);
                    set_term_key(cform,xch);
                    return(_winfo.errno=W_ESCPRESS); 
                }
                break;

            case 0x1c0a:    /* Ctrl-Enter */
                set_term_key(cform,xch);
                goto form_done;

            case 0x1c0d:    /* Enter */

                /* abbreviate pointer to current field */
                cfield=cform->cfield;

                /* if field is numeric, then format it so */
                if(cfield->fconv=='9')
                    deci_right(cform,((cform->pbuf==cfield->buf)) ?
                      (cfield->buf+cfield->lenbuf) : cform->pbuf);

                /* if we are at last field, validate it, */
                /* then copy temp buffers to receiving   */
                /* buffers, free input records, and exit */
                if(cform->cfield==last_rec(cform)) {
                    set_term_key(cform,xch);
form_done:
                    if(validate_field(cform)) break;
                    after_edit(cform);
                    cform->cfield=cform->field;
                    while(cform->cfield!=NULL) {
                        strcpy(cform->cfield->str,cform->cfield->buf);
                        cform->cfield=cform->cfield->prev;
                    }
                    free_fields(cform);
                    return(_winfo.errno=W_NOERROR);  
                }
                /* fall through to Tab */

            case 0x0f09:    /* Tab */

                /* if field is valid, display it, go      */
                /* to next field, first position. if      */
                /* conditional update feature is on, then */
                /* position cursor at end of line         */
                if(!goto_field(cform,FLD_NR)) {
                    beg_pos(cform);
                    if((edit=cond_update(cform))==NO) continue;
                }
                break;

            case 0x0e08:    /* BackSpace */

                /* if currently at 1st position of 1st field, ignore. */
                /* otherwise, display a destructive backspace         */
                if(!( (cform->pbuf==cform->cfield->buf) &&
                  (cform->cfield==first_rec(cform)) )) back_space(cform);
                break;

            case 0x4700:    /* Home */

                /* go to first position in field */
                beg_pos(cform);
                break;

            case 0x0f00:    /* Shift-Tab */

                /* if field is valid, display it, go to   */
                /* previous field, first position.  if    */
                /* conditional update feature is on, then */
                /* position cursor at end of line         */
                if(!goto_field(cform,FLD_PR)) {
                    beg_pos(cform);
                    if((edit=cond_update(cform))==NO) continue;
                }
                break;

            case 0x4800:    /* UpArrow */

                /* if field is valid, display it, and */
                /* move cursor up to previous field   */
                goto_field(cform,FLD_UP);
                break;

            case 0x4b00:    /* LeftArrow */

                /* go to previous position in field */
                prev_pos(cform);
                break;

            case 0x4d00:    /* RightArrow */

                /* go to next position in field */
                next_pos(cform);
                break;

            case 0x4f00:    /* End */

                /* go to last position in field */
                end_line(cform);
                break;

            case 0x5000:    /* DownArrow */

                /* if field is valid, display it, and */
                /* move cursor down to next field     */
                goto_field(cform,FLD_DN);
                break;

            case 0x5200:    /* Ins */

                /* toggle Insert mode */
                if(cform->insert) {
                    cform->insert=OFF;
                    smcursor();
                }
                else {
                    cform->insert=ON;
                    lgcursor();
                }
                break;

            case 0x5300:    /* Del */

                /* delete character under cursor */
                del_char(cform);
                break;

            case 0x7500:    /* Ctrl-End */

                /* if field is valid, display it, then */
                /* go to last position of last field   */
                if(!goto_field(cform,FLD_LS)) end_pos(cform);
                break;

            case 0x7700:    /* Ctrl-Home */

                /* if field is valid, display it, then */
                /* go to first position of first field */
                if(!goto_field(cform,FLD_FR)) beg_pos(cform);
                break;

            case 0x0e7f:    /* Ctrl-BackSpace */

                /* delete word to left of cursor */
                del_left_word(cform);
                break;

            case 0x1312:    /* Ctrl-R */

                /* restore field to original contents */
                restore_field(cform,1);
                break;

            case 0x1414:    /* Ctrl-T */

                /* delete word to right of cursor */
                del_word(cform);
                break;

            case 0x1615:    /* Ctrl-U */

                /* delete from cursor to end of field */
                del_field(cform,cform->pbuf,1);
                break;

            case 0x1519:    /* Ctrl-Y */

                /* delete from cursor to end of last field */
                del_rest(cform);
                break;

            case 0x7300:    /* Ctrl-LeftArrow */

                /* move cursor to previous word to the left */
                prev_word(cform);
                break;

            case 0x7400:    /* Ctrl-RightArrow */

                /* move cursor to next word to the right */
                next_word(cform);
                break;

            case 0x342e:    /* Period */
            case 0x532e:    /* Shift-Del (decimal) */

                /* if field is numeric, right justify, space fill */
                /* digits to the left of the decimal point, and   */
                /* move cursor to position after decimal point    */
                if(cform->cfield->fconv=='9') {
                    deci_left(cform,cform->pbuf);
                    disp_field(cform,NULL,1);
                    temp=cform->cfield;
                    while(!cform->decimal&&cform->cfield==temp)
                        next_pos(cform);
                    break;
                }
                /* fall through to default */

            default:

                /* if conditional update feature is on, check to see if */
                /* an editing key has been pressed.  if it has not,     */
                /* then erase field and position cursor at beginning    */
                if(cform->cfield->mode==2) {
                    if(!edit) {
                        beg_pos(cform);
                        del_field(cform,cform->pbuf,1);
                    }
                }

                /* break keycode down to just the ASCII code portion. if */
                /* ASCII code is zero, then it must be an extended key   */
                ch=(char)xch;
                if(!isprint(ch)) break;

                /* if lower, mixed, or upper case conversion */
                /* was specified, then do the conversion     */
                if(*cform->pformat=='M' || cform->cfield->fconv=='M')
                    ch=touplow(cform->cfield->buf,cform->pbuf,ch);
                else if(*cform->pformat=='U' || cform->cfield->fconv=='U')
                    ch=toupper(ch);
                else if(*cform->pformat=='L' || cform->cfield->fconv=='L')
                    ch=tolower(ch);

                /* see if current format string character is the left angle  */
                /* bracket (start of set inclusion).  if so, search through  */
                /* the set of characters for the character matching          */
                /* keypress.  if a matching character wasn't found, then the */
                /* input character is invalid.                               */
                if(*cform->pformat=='<') {
                    valid=NO;
                    p=cform->pformat;
                    cform->pformat++;
                    while(*cform->pformat!='>') {
                        if(ch==*cform->pformat) valid=YES;
                        cform->pformat++;
                    }
                    if(!valid) cform->pformat=p;
                }

                /* see if current format string character is the left square */
                /* bracket (start of set exclusion).  if so, search through  */
                /* the set of characters for the character matching          */
                /* keypress.  if a matching character wasn't found, then the */
                /* input character is valid.                                 */
                else {
                    if(*cform->pformat=='[') {
                        valid=YES;
                        p=cform->pformat;
                        cform->pformat++;
                        while(*cform->pformat!=']') {
                            if(ch==*cform->pformat) valid=NO;
                            cform->pformat++;
                        }
                        if(!valid) cform->pformat=p;
                    }

                    /* see if input character is a valid character */
                    /* of the current format character type        */
                    else {
                        valid=cvaltype(ch,*cform->pformat);
                    }
                }

                /* if input character is a space, then even */
                /* if it's not valid, make it valid for now */
                if(ch==' ') valid=YES;

                /* if input character is valid, display it */
                if(valid) {

                    /* if at last position of last field, don't wrap-around */
                    if((cform->pbuf==(cform->cfield->buf+cform->cfield->lenbuf
                      -1))&&(cform->cfield==last_rec(cform))) {
                        valid=NO;
                        end_pos(cform);
                    }
                    temp=cform->cfield;
                    disp_char(cform,ch,valid);
                    if(!valid&&cform->cfield->fconv=='9') beg_pos(cform);
                    if(temp!=cform->cfield)
                        if((edit=cond_update(cform))==NO) continue;

                    /* if field is numeric, and we just passed the decimal  */
                    /* point, right justify, space fill the left side of it */
                    if(valid) {
                        if(cform->cfield->fconv=='9'&&cform->decimal) {
                            deci_left(cform,cform->cfield->buf+
                                      cform->cfield->decpos-1);
                            disp_field(cform,NULL,1);
                        }
                    }
                }
        }

        /* turn on edit mode */
        edit=YES;
    }
}

/*---------------------------------------------------------------------------*/

/* this function will start the cursor at the 1st position in */
/* the field and advance it until we are at specified column  */

static void adv_column(struct _form_t *cform,int col)
{
    struct _field_t *temp;

    temp=cform->cfield;
    beg_pos(cform);
    while(cform->cwcol<col) next_pos(cform);
    if(cform->cfield!=temp) while(cform->cfield!=temp) prev_pos(cform);
}

/*---------------------------------------------------------------------------*/

/* this function takes care of a couple tasks to do after editing a field */

static void after_edit(struct _form_t *cform)
{
    /* redisplay current field in field attribute */
    disp_field(cform,NULL,0);

    /* if an 'after' function was defined, call it */
    call_func(cform->cfield->after);

    /* turn off Insert mode */
    cform->insert=OFF;
    smcursor();
}

/*---------------------------------------------------------------------------*/

/* this function does a destructive backspace */

static void back_space(struct _form_t *cform)
{
    /* decrement buffer pointer */
    cform->pbuf--;

    /* if buffer pointer is less than start of buffer, then validate */
    /* field, display field, and go to previous field, end of line   */
    /* otherwise, decrement column, format string */
    /* pointer, and back up one position          */
    if( cform->pbuf < cform->cfield->buf ) {
        if(!goto_field(cform,FLD_PR)) end_line(cform);
    }
    else {
        cform->cwcol--;
        cform->pformat--;
        prev_fchar(cform);
    }

    /* if insert mode is on, delete character under cursor */
    /* otherwise, replace character with a space           */
    if(cform->insert)
        del_char(cform);
    else {
        wprintc(cform->cwrow,cform->cwcol,cform->textattr,' ');
        *cform->pbuf=' ';
    }
}

/*---------------------------------------------------------------------------*/

/* this function sets current position to the beginning of the field */

static void beg_pos(struct _form_t *cform)
{
    struct _field_t *cfield;

    cfield         = cform->cfield;
    cform->cwrow   = cfield->wrow;
    cform->cwcol   = cfield->wcol;
    cform->pbuf    = cfield->buf;
    cform->pformat = cfield->format;
    next_fchar(cform);
}

/*---------------------------------------------------------------------------*/

/* this function calls the given function, then checks all */
/* fields to see if any of their redisplay flags have been */
/* set.  If so, then the field will be redisplayed.        */

static void call_func(void (*func)(void))
{
    register int err;

    if(func!=NULL) {
        err=whelpush();
        (*func)();
        if(!err) whelpop();
        check_redisp(_winfo.active->form);
    }
}

/*---------------------------------------------------------------------------*/

/* this function checks syntax of format string, returns length of field */

static void check_format(char *format,int *lenfld,int *lenbuf,int *decpos)
{
    char *p,ch;
    int valid=YES,deccount=0;

    *lenbuf=0;
    *lenfld=0;
    *decpos=0;
    p=format;

    /* do while not end of string and string still valid */
    while(*p && valid) {    

        /* test current character */
        switch(*p) {

            case '\"':
            case '\'':

                /* search for matching quote */
                ch=*p++;
                while(*p!=ch) {

                    /* check for premature end of string */
                    if(!*p) {
                        valid=NO;
                        break;
                    }
                    (*lenfld)++;
                    p++;
                }
                p++;
                break;

            case ' ':

                /* skip spaces, they are used for readability */
                p++;
                break;

            case '<':   /* left angle bracket (start of inclusion set) */

                /* search for matching angle bracket */
                p++;
                while(*p!='>') {

                    /* check for premature end of string */
                    if(!*p) {
                        valid=NO;
                        break;
                    }
                    p++;
                }
                (*lenfld)++;
                (*lenbuf)++;
                p++;
                break;

            case '[':   /* left square bracket (start of exclusion set) */

                /* search for matching square bracket */
                p++;
                while(*p!=']') {

                    /* check for premature end of string */
                    if(!*p) {
                        valid=NO;
                        break;
                    }
                    p++;
                }
                (*lenfld)++;
                (*lenbuf)++;
                p++;
                break;

            case '.':   /* decimal point */

                /* mark decimal position */
                *decpos=(*lenbuf)+1;
                (*lenfld)++;
                p++;

                /* make sure there is only 1 decimal point */
                deccount++;
                if(deccount>1) valid=NO;
                break;

            default:

                /* see if character is a valid format character type */
                if( cvaltype(0,*p) == -1 ) {
                    valid=NO;
                    break;
                }
                (*lenfld)++;
                (*lenbuf)++;
                p++;
        }
    }

    /* if no decimal was found, "position" it at end of field */
    if(!(*decpos)) *decpos=(*lenbuf)+1;
    if(!valid) {
        *lenbuf=0;
        *lenfld=0;
        *decpos=0;
    }
}

/*---------------------------------------------------------------------------*/

static void check_redisp(struct _form_t *cform)
{
    struct _field_t *test,*cfield;

    /* search all fields in linked list for fields who have had */
    /* their redisplay flag set.  Redisplay all thos that have. */
    for(test=cform->field;test!=NULL;test=test->prev) {
        if(test->redisp) {
            cfield=cform->cfield;
            cform->cfield=test;
            disp_field(cform,NULL,test==cfield?1:0);
            cform->cfield=cfield;
            test->redisp=OFF;
        }
    }
}

/*---------------------------------------------------------------------------*/

/* this function returns the given field's window */
/* column closest to the current window column    */

static int closest_column(struct _form_t *cform,struct _field_t *field)
{
    register int cwcol,wcol;
    int tcol,lcol;

    cwcol=cform->cwcol;
    wcol=field->wcol;
    lcol=wcol+field->lenfld-1;
    if( (cwcol>=wcol) && (cwcol<=lcol) )
        tcol=cwcol;
    else
        if(wcol<cwcol)
            tcol=lcol;
        else
            tcol=wcol;
    return(tcol);
}

/*---------------------------------------------------------------------------*/

/* this function prepares for conditional update */

static int cond_update(struct _form_t *cform)
{
    /* if conditional update, then start cursor end of line */
    if(cform->cfield->mode==2) {
        if(!strblank(cform->pbuf)) end_line(cform);
        return(NO);
    }
    return(YES);
}

/*---------------------------------------------------------------------------*/

/* this function will right justify, space fill numbers before the decimal */

static void deci_left(struct _form_t *cform,char *start)
{
    char *p,*q,*left;
    struct _field_t *cfield;

    /* abbreviate pointer to current field */
    cfield=cform->cfield;

    left=cfield->buf+cfield->decpos-2;

    /* blank out from starting position to left side of decimal */
    for(p=start;p<=left;*p++=' ');

    /* search for first space to left of decimal */
    for(p=left;*p!=' '&&p>cfield->buf;p--);

    if(p>cfield->buf) {

        /* while q >= first position in field */
        q=p;
        while(q>=cfield->buf) {
            for(;(*q==' ')&&(q>=cfield->buf);q--);
            while((q>=cfield->buf)&&(*q!=' ')) {
                *p--=*q;
                *q--=' ';
            }
        }
    }

    /* change leading zeros to blanks */
    q=left;
    for(p=cfield->buf;*p==' '&&p<q;p++);
    if(p<q) while(*p=='0'&&p<q) *p++=' ';

    /* check for all blanks to left of decimal */
    if(*left==' ') *left='0';
}

/*---------------------------------------------------------------------------*/

/* this function will right justify, space fill numbers before the decimal
   point, and left justify, zero fill numbers after the decimal point */

static void deci_right(struct _form_t *cform,char *start)
{
    char *p,*q,*last,*right;
    struct _field_t *cfield;

    /* abbreviate pointer to current field */
    cfield=cform->cfield;

    /* calculate last position and position to right of decimal */
    last=cfield->buf+cfield->lenbuf-1;
    right=cfield->buf+cfield->decpos-1;

    /* blank out from start position to end of field */
    for(p=start;p<=last;*p++=' ');

    /* first do left side of decimal */
    deci_left(cform,start);

    /* search for first space to right of decimal */
    for(p=right;*p!=' '&&p<last;p++);

    if(p<last) {

        /* while q <= last position in field */
        q=p;
        while(q<=last) {
            for(;*q==' '&&q<=last;q++);
            while(q<=last&&*q!=' ') {
                *p++=*q;
                *q++=' ';
            }
        }
    }

    /* change trailing blanks to zeros */
    for(p=last;*p==' '&&p>=right;p--) *p='0';
}

/*---------------------------------------------------------------------------*/

/* this function deletes a character and shifts following text left */

static void del_char(struct _form_t *cform)
{
    /* shift characters in buffer left 1 position */
    strshl(cform->pbuf,1);

    /* re-display field */
    disp_field(cform,cform->pbuf,1);
}

/*---------------------------------------------------------------------------*/

/* this function is used to delete characters to the end of the field */

static void del_field(struct _form_t *cform,char *start,int high)
{
    /* blank out field from start position */
    strset(start,' ');

    /* re-display field */
    disp_field(cform,start,high);
}

/*---------------------------------------------------------------------------*/

/* this functions deletes the word to the left */

static void del_left_word(struct _form_t *cform)
{
    char *buf;
    int back=0;

    /* abbreviate pointer to start of field's buffer */
    buf=cform->cfield->buf;

    /* see if at 1st position in field.  */
    /* if so, just do a normal backspace */
    if(cform->pbuf==buf) {
        back_space(cform);
        return;
    }

    /* start 1 position back */
    prev_pos(cform);
    back++;

    /* skip over any spaces */
    while(cform->pbuf>buf&&*cform->pbuf==' ') {
        prev_pos(cform);
        back++;
    }

    /* find beginning of word under cursor */
    while(cform->pbuf>buf&&*cform->pbuf!=' ') {
        *cform->pbuf=' ';
        prev_pos(cform);
        back++;
    }
    if(cform->pbuf==buf) {
        *cform->pbuf=' ';
    }
    else {
        next_pos(cform);
        back--;
    }

    /* skip over any spaces */
    while(cform->pbuf>buf&&*cform->pbuf==' ') {
        prev_pos(cform);
        back++;
    }
    if(*cform->pbuf!=' ') {
        next_pos(cform);
        back--;
    }

    /* if Insert mode is on, shift text after cursor to the left */
    if(cform->insert&&cform->pbuf>=buf) strshl(cform->pbuf,back);

    /* re-display field */
    disp_field(cform,cform->pbuf,1);
}

/*---------------------------------------------------------------------------*/

/* this function deletes from current position through rest of the fields */

static void del_rest(struct _form_t *cform)
{
    register struct _field_t *temp;

    /* save pointer to current field */
    temp=cform->cfield;

    /* delete to end of current field */
    del_field(cform,cform->pbuf,1);

    /* while not last field, find next field, and erase it */
    while(cform->cfield!=last_rec(cform)) {
        cform->cfield=next_rec(cform);
        del_field(cform,cform->cfield->buf,0);
    }

    /* restore pointer to current field */
    cform->cfield=temp;
}

/*---------------------------------------------------------------------------*/

/* this function deletes the word to the right of the cursor */

static void del_word(struct _form_t *cform)
{
    register char *p,*q;

    /* find end of word under cursor */
    p=q=cform->pbuf;
    while(*q && *q!=' ') q++;

    /* find end of spaces */
    while(*q==' ') q++;

    /* move characters from right to left */
    while(*q) *p++=*q++;

    /* delete trailing characters */
    if(*p) del_field(cform,p,1);

    /* display the field */
    disp_field(cform,cform->pbuf,1);
}

/*---------------------------------------------------------------------------*/

/* this function displays a character and advances to next position */

static void disp_char(struct _form_t *cform,int ch,int advance)
{
    struct _field_t *temp;

    temp=cform->cfield;
    if(cform->insert) insert_char(cform);
    wprintc(cform->cwrow,cform->cwcol,cform->textattr,
            (*cform->pformat=='P'||cform->cfield->fconv=='P')?' ':ch);
    *cform->pbuf=ch;
    if(advance) {
        cform->pbuf++;
        if( cform->pbuf >= (cform->cfield->buf+cform->cfield->lenbuf) ) {
            if(!goto_field(cform,FLD_NR)) beg_pos(cform);
        }
        else {
            cform->cwcol++;
            cform->pformat++;
            next_fchar(cform);
        }
        if(temp!=cform->cfield&&cform->cfield->mode==2&&!strblank(cform->pbuf))
            end_line(cform);
    }
}

/*---------------------------------------------------------------------------*/

/* this function displays the field of the specified window record */

static void disp_field(struct _form_t *cform,char *start,int high)
{
    register struct _field_t *cfield;
    register int ccol;
    int crow,attr,disp;
    char *pf,*pb,ch;

    /* initialize position */
    cfield = cform->cfield;
    crow   = cfield->wrow;
    ccol   = cfield->wcol;
    pb     = cfield->buf;
    pf     = cfield->format;

    /* set starting point */
    if(start==NULL) start=cfield->buf;

    while(*pf) {

        /* select character attribute depending on input hilite flag */
        attr=high?cform->textattr:cform->fieldattr;

        /* see if we are at starting point yet */
        disp=(pb>=start)?YES:NO;

        /* test current format character */
        switch(*pf) {

            case '\"':
            case '\'':

                /* display all text inside quotes */
                ch=*pf++;
                while(*pf!=ch) {
                    if(disp) wprintc(crow,ccol,attr,*pf);
                    ccol++;
                    pf++;
                }
                break;
            
            case ' ':

                /* skip readability spaces */
                break;

            case '<':

                /* find right angle bracket, and    */
                /* display current buffer character */
                while(*pf!='>') pf++;
                if(disp) wprintc(crow,ccol,attr,*pb);
                ccol++;
                pb++;
                break;

            case '[':

                /* find right square bracket, and   */
                /* display current buffer character */
                while(*pf!=']') pf++;
                if(disp) wprintc(crow,ccol,attr,*pb);
                ccol++;
                pb++;
                break;

            case '.':

                /* display decimal point */
                if(disp) wprintc(crow,ccol,attr,*pf);
                ccol++;
                break;

            default:

                /* if current format character is a password character, */
                /* or field is a password field, then display a blank.  */
                /* otherwise, display the current buffer character.     */
                if(disp)
                    wprintc(crow,ccol,attr,(*pf=='P'||cfield->fconv=='P')
                      ?' ':*pb);
                ccol++;
                pb++;
        }
        pf++;
    }
}

/*---------------------------------------------------------------------------*/

/* this function sets the cursor at next field down from current field */

static void down_field(struct _form_t *cform)
{
    struct _field_t *best,*temp;
    int tcol,brow,bcol,tdist,bdist,trow,crow,ccol;

    /* initialize best record to NULL */
    best = NULL;
    brow = bcol = 32767;

    /* find the closest field downwards */
    crow=(int)cform->cwrow;
    ccol=(int)cform->cwcol;
    for(temp=cform->field;temp!=NULL;temp=temp->prev) {
        trow=(int)temp->wrow;
        tcol=closest_column(cform,temp);
        if(trow>crow) {
            tdist=abs(ccol-tcol);
            bdist=abs(ccol-bcol);
            if((trow<brow)||((trow==brow&&tdist<bdist))) {
                best  = temp;
                brow  = trow;
                bcol  = tcol;
            }
        }
    }

    /* if a field was found, then set current field to found */
    /* field, and advance to the matching column position    */
    if(best!=NULL) {
        cform->cfield=best;
        adv_column(cform,bcol);
    }
}

/*---------------------------------------------------------------------------*/

/* this function sets current position to end of line in field */

static void end_line(struct _form_t *cform)
{
    char *temp;

    /* save current position */
    temp=cform->pbuf;

    /* initialize to last position in field */
    end_pos(cform);

    /* if field is empty, stay on last position of field */
    if(strblank(cform->cfield->buf) || *cform->pbuf!=' ') return;

    /* search backwards for non-space or beginning of field */
    while(*cform->pbuf==' ') {
        if(cform->pbuf<=cform->cfield->buf) {
            end_pos(cform);
            return;
        }
        prev_pos(cform);
    }
    next_pos(cform);

    /* if position didn't change go to end of field */
    if(temp==cform->pbuf) end_pos(cform);
}

/*---------------------------------------------------------------------------*/

/* this function sets current position to the end of the field */

static void end_pos(struct _form_t *cform)
{
    struct _field_t *cfield;

    cfield         = cform->cfield;
    cform->cwrow   = cfield->wrow;
    cform->cwcol   = cfield->wcol   + cfield->lenfld - 1;
    cform->pbuf    = cfield->buf    + cfield->lenbuf - 1;
    cform->pformat = cfield->format + cfield->lenfmt - 1;
    prev_fchar(cform);
}

/*---------------------------------------------------------------------------*/

/* this function finds the first defined field */

static struct _field_t *first_rec(struct _form_t *cform)
{
    struct _field_t *temp;

    for(temp=cform->field;temp->prev!=NULL;temp=temp->prev) ;
    return(temp);
}

/*---------------------------------------------------------------------------*/

/* this function frees memory held by all of the input definitions */

static void free_fields(struct _form_t *cform)
{
    struct _form_t  *form;
    struct _field_t *field;

    /* do while more fields */
    while(cform->field!=NULL) {

        /* free temp string buffer */
        free(cform->field->buf);

        /* free input field record and update linked list */
        field=cform->field->prev;
        free(cform->field);
        cform->field=field;
        if(cform->field!=NULL) cform->field->next=NULL;
    }

    /* free input form record and update linked list */
    form=cform->prev;
    free(cform);
    _winfo.active->form=form;
    if(_winfo.active->form!=NULL) _winfo.active->form->next=NULL;
}

/*---------------------------------------------------------------------------*/

/* this function will validate the current field, then move the cursor */
/* to a new field, calling the after and before functions if defined.  */

static int goto_field(struct _form_t *cform,int which)
{
    int err;

    /* validate current field */
    if((err=validate_field(cform))!=0) return(err);

    /* display field, call 'after' function, and turn off Insert mode */
    after_edit(cform);

    /* call function which does the move to the new field */
    if(which==FLD_UP)       up_field(cform);
    else if(which==FLD_DN)  down_field(cform);
    else                    cform->cfield=(*funcs[which])(cform);

    /* set help category */
    _winfo.help=cform->cfield->help;

    /* display field in current field attribute */
    disp_field(cform,NULL,1);

    /* if a 'before' function was defined, call it */
    call_func(cform->cfield->before);

    /* return normally */
    return(0);
}

/*---------------------------------------------------------------------------*/

/* this function inserts a character, shifting text right one position */

static void insert_char(struct _form_t *cform)
{
    /* shift characters right 1 place starting with current position */
    strshr(cform->pbuf,1);

    /* re-display the field */
    disp_field(cform,cform->pbuf,1);
}

/*---------------------------------------------------------------------------*/

/* this function finds the last defined field */

static struct _field_t *last_rec(struct _form_t *cform)
{
    return(cform->field);
}

/*---------------------------------------------------------------------------*/

/* this function finds next non-text format character in format string */

static void next_fchar(struct _form_t *cform)
{
    char ch;
    int done=NO;

    cform->decimal=NO;

    /* do while no format character */
    while(!done) {

        /* test current character in format string */
        switch(*cform->pformat) {

            case '\"':
            case '\'':

                /* find matching quote */
                ch=*cform->pformat++;
                while(ch!=*cform->pformat) {
                    cform->cwcol++;
                    cform->pformat++;
                }
                cform->pformat++;
                break;

            default:

                /* if currrently inside an angle or square */
                /* bracket list, find end of list          */
                if(*(cform->pformat-1)=='<') {
                    while(*cform->pformat!='>') cform->pformat++;
                    cform->pformat++;
                }
                else {
                    if(*(cform->pformat-1)=='[') {
                        while(*cform->pformat!=']') cform->pformat++;
                        cform->pformat++;
                    }
                    else {

                        /* skip readability spaces */
                        if(*cform->pformat==' ') {
                            cform->pformat++;
                        }
                        else {

                            /* check for decimal point */
                            if(*cform->pformat=='.') {
                                cform->pformat++;
                                cform->cwcol++;
                                cform->decimal=YES;
                            }
                            else {

                                /* must be at next format character */
                                done=YES;
                            }
                        }
                    }
                }
        }
    }
}

/*---------------------------------------------------------------------------*/

/* this function advances the cursor to the next position in field */

static void next_pos(struct _form_t *cform)
{
    /* increment buffer pointer */
    cform->pbuf++;

    /* if buffer pointer is greater than last position in field     */
    /* and field is valid, then display field and go to next field, */
    /* first position. otherwise, advance to next position          */
    if( cform->pbuf >= (cform->cfield->buf+cform->cfield->lenbuf) ) {
        if(!goto_field(cform,FLD_NR)) beg_pos(cform);
    }
    else {
        cform->cwcol++;
        cform->pformat++;
        next_fchar(cform);
    }
}

/*---------------------------------------------------------------------------*/

/* this function sets the current field to the next field in the linked list */

static struct _field_t *next_rec(struct _form_t *cform)
{
    struct _field_t *field;

    field=cform->cfield->next;
    if(field==NULL) field=first_rec(cform);
    return(field);
}

/*---------------------------------------------------------------------------*/

/* this function advances the cursor to the next word */

static void next_word(struct _form_t *cform)
{
    next_pos(cform);
    while(*cform->pbuf!=' ' &&
          cform->pbuf!=(cform->cfield->buf+cform->cfield->lenbuf-1))
        next_pos(cform);
    while(*cform->pbuf==' ' &&
          cform->pbuf!=(cform->cfield->buf+cform->cfield->lenbuf-1))
        next_pos(cform);
    if(cform->pbuf==(cform->cfield->buf+cform->cfield->lenbuf-1))
        next_pos(cform);
}

/*---------------------------------------------------------------------------*/

/* this function finds previous non-text format character in format string */

static void prev_fchar(struct _form_t *cform)
{
    char ch;
    int done=NO;

    /* do while format character not found */
    while(!done) {

        /* test current format character */
        switch(*cform->pformat) {

            case '\"':
            case '\'':

                /* find matching quote */
                ch=*cform->pformat--;
                while(ch!=*cform->pformat) {
                    cform->cwcol--;
                    cform->pformat--;
                }
                cform->pformat--;
                break;

            case ' ':

                /* skip readability spaces */
                cform->pformat--;
                break;

            case '>':

                /* find left angle bracket */
                while(*cform->pformat!='<') cform->pformat--;
                done=YES;
                break;

            case ']':

                /* find left square bracket */
                while(*cform->pformat!='[') cform->pformat--;
                done=YES;
                break;

            case '.':

                /* skip decimal point   */
                cform->pformat--;
                cform->cwcol--;
                break;

            default:

                /* must be previous format character */
                done=YES;
        }
    }
}

/*---------------------------------------------------------------------------*/

/* this function positions the cursor to the previous position in field */

static void prev_pos(struct _form_t *cform)
{
    /* decrement buffer pointer */
    cform->pbuf--;

    /* if buffer pointer is less than first position in field and   */
    /* field is valid, then display field and go to previous field, */
    /* last position. otherwise, advance to next position           */
    if( cform->pbuf < cform->cfield->buf ) {
        if(!goto_field(cform,FLD_PR)) end_pos(cform);
    }
    else {
        cform->cwcol--;
        cform->pformat--;
        prev_fchar(cform);
    }
}

/*---------------------------------------------------------------------------*/

/* this function sets the current field to the prev field in linked list */

static struct _field_t *prev_rec(struct _form_t *cform)
{
    struct _field_t *field;

    field=cform->cfield->prev;
    if(field==NULL) field=last_rec(cform);
    return(field);
}

/*---------------------------------------------------------------------------*/

/* this function moves the cursor to the beginning of the previous word */

static void prev_word(struct _form_t *cform)
{
    prev_pos(cform);
    while(*cform->pbuf==' ' && cform->pbuf!=cform->cfield->buf)
        prev_pos(cform);
    while(*cform->pbuf!=' ' && cform->pbuf!=cform->cfield->buf)
        prev_pos(cform);
    if(cform->pbuf!=cform->cfield->buf) next_pos(cform);
}

/*---------------------------------------------------------------------------*/

/* this function restores a field to its contents before editing */

static void restore_field(struct _form_t *cform,int high)
{
    struct _field_t *cfield;

    /* abbreviate pointer to current field */
    cfield=cform->cfield;

    strcpy(cfield->buf,cfield->str);
    if(cfield->fconv=='9') deci_right(cform,cfield->buf+cfield->lenbuf);
    disp_field(cform,NULL,high);
}

/*---------------------------------------------------------------------------*/

/* this function sets the termination key variable, if defined */

static void set_term_key(struct _form_t *cform,unsigned xch)
{
    if(cform->termkey!=NULL) *(cform->termkey)=xch;
}

/*---------------------------------------------------------------------------*/

/* this function searches for the previous field up from current field */

static void up_field(struct _form_t *cform)
{
    struct _field_t *best,*temp;
    int brow,bcol,trow,tcol,tdist,bdist,crow,ccol;

    /* initialize best record to NULL */
    best  = NULL;
    brow  = -1;
    bcol  = 32767;

    /* find the closest field upwards */
    crow=(int)cform->cwrow;
    ccol=(int)cform->cwcol;
    for(temp=cform->field;temp!=NULL;temp=temp->prev) {
        trow=(int)temp->wrow;
        tcol=closest_column(cform,temp);
        if(trow<crow) {
            tdist=abs(ccol-tcol);
            bdist=abs(ccol-bcol);
            if((trow>brow)||((trow==brow&&tdist<bdist))) {
                best=temp;
                brow=trow;
                bcol=tcol;
            }
        }
    }

    /* if a field was found, then set current field to found */
    /* field, and advance to the matching column position    */
    if(best!=NULL) {
        cform->cfield=best;
        adv_column(cform,bcol);
    }
}

/*---------------------------------------------------------------------------*/

/* this function validates the field buffer using the format string */

static int validate_field(struct _form_t *cform)
{
    char *pb,*pf,ch;
    struct _field_t *cfield;
    struct _form_t  *temp;
    int valid=YES,i,pos;

    /* abbreviate pointer to current field */
    cfield=cform->cfield;

    /* if field is numeric, then format it so */
    if(cfield->fconv=='9') deci_right(cform,cfield->buf+cfield->lenbuf);

    pb=cfield->buf;
    pf=cfield->format;

    /* if string is blank, don't do the built-in validation. */
    /* As far as we're concerned, a blank field is okay.     */
    if(!strblank(pb)) {

        /* do while not end of format string and field still valid */
        while(*pf && valid) {

            /* test character in the field buffer with */
            /* current format character for validity.  */
            switch(*pf) {

                case ' ':

                    /* skip readability spaces */
                    pf++;
                    break;

                case '\"':
                case '\'':

                    /* skip quoted text in format string */
                    ch=*pf++;
                    while(ch!=*pf) pf++;
                    pf++;
                    break;

                case '<':

                    /* check current character in field buffer for */
                    /* inclusion in the list of valid characters.  */
                    pf++;
                    valid=NO;
                    while(*pf!='>') {
                        if(*pb==*pf) valid=YES;
                        pf++;
                    }
                    if(valid) {
                        pf++;
                        pb++;
                    }
                    break;

                case '[':


                    /* check current character in field buffer for */
                    /* exclusion in the list of valid characters.  */
                    pf++;
                    valid=YES;
                    while(*pf!=']') {
                        if(*pb==*pf) valid=NO;
                        pf++;
                    }
                    if(valid) {
                        pf++;
                        pb++;
                    }
                    break;

                case '.':

                    /* skip decimal point */
                    pf++;
                    break;

                default:

                    /* see if current field character is a valid */
                    /* type of the current format character      */
                    valid=cvaltype(*pb,*pf);
                    if(*pb==' '&&cfield->fconv=='9') valid=YES;
                    if(valid) {
                        pf++;
                        pb++;
                    }
            }
        }
    }

    /* if the field passes validation so far and a user-defined */
    /* function was specified, save current environment, call   */
    /* the user function, then restore the environment          */
    if(valid) {
        if(cfield->validate!=NULL) {
            temp=cform;
            pos=( (*(cform->cfield->validate)) (cform->cfield->buf) );
            cform=temp;
            cfield=cform->cfield;

            /* check to see if any of the fields' redisp flags have been set */
            check_redisp(cform);

            /* test return value from user-defined function. */
            /* if not zero, then field is in error           */
            if(pos) {
                beg_pos(cform);
                goto validate_error;
            }
        }
    }
    else {

        /* advance to error position in field */
        beg_pos(cform);
        pos=(int)( ((unsigned long) pb ) - ((unsigned long) cfield->buf ) + 1);
validate_error:
        if(pos>cfield->lenbuf) pos=1;
        for(i=1;i<pos;i++) next_pos(cform);

        /* return error code to caller */
        return(pos);
    }

    /* return with no error (field is valid) */
    return(0);
}
