/*  CXL - Copyright (c) 1987-1989 by Mike Smedley - All Rights Reserved  */
/*  WMENU.C  -  wmenubeg()  starts a menu definition
                wmenuitem() defines a menu item
                wmenuend()  ends a menu definition
                wmenuget()  processes the defined menu      */

#include <conio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "cxldef.h"
#include "cxlkey.h"
#include "cxlmou.h"
#include "cxlvid.h"
#include "cxlwin.h"

/* prototypes for local functions */
static int              calc_bar_width(struct _menu_t *wmenu,
                                       struct _item_t *witem);
static int              calc_center_item(struct _item_t *item);
static void             call_after(struct _item_t *citem);
static void             call_before(struct _item_t *citem);
static void             call_func(void (*func)(void));
static void             close_window(WINDOW w);
static void             disp_item(struct _item_t *witem,int bar);
static struct _item_t  *down_item(struct _item_t *curr);
static struct _item_t  *first_item(void);
static void             free_menu(struct _menu_t *wmenu);
static struct _item_t  *goto_item(struct _item_t *citem,int which);
static void             hide_mouse_cursor(void);
static struct _item_t  *last_item(void);
static struct _item_t  *left_item(struct _item_t *curr);
static struct _item_t  *mouse_on_item(struct _menu_t *menu,int mrow,int mcol);
static void             post_move(struct _item_t *citem);
static void             pre_exit(WINDOW w,int close);
static void             pre_move(struct _item_t *citem);
static int              read_mouse(struct _item_t *citem);
static struct _item_t  *right_item(struct _item_t *curr);
static void             show_mouse_cursor(void);
static struct _item_t  *up_item(struct _item_t *curr);

/*---------------------------------------------------------------------------*/

/* menu item movement definitions */
#define ITM_LT  0
#define ITM_RT  1
#define ITM_UP  2
#define ITM_DN  3
#define ITM_FR  4
#define ITM_LS  5

/*---------------------------------------------------------------------------*/

/* array of function pointers to some of the item movement functions */
static struct _item_t *(*funcs[4])(struct _item_t *) =
    { left_item,right_item,up_item,down_item };

/* flag used for initial display of menu selections */
static int _Near _Cdecl dispdesc=YES;

/*---------------------------------------------------------------------------*/

int wmenubeg(int srow,int scol,int erow,int ecol,int btype,int battr,
             int wattr,void (*open)(void))
{
    struct _menu_t *wmenu;

    /* if this is a submenu, then make sure that it is under a menu item */
    if(_winfo.mlevel>_winfo.ilevel) return(_winfo.errno=W_NOITMDEF);

    /* allocate memory for new menu record */
    if((wmenu=malloc(sizeof(struct _menu_t)))==NULL)
        return(_winfo.errno=W_ALLOCERR);

    /* if menu is not a submenu, make it a new base menu record */
    if(!_winfo.mlevel) {
        if(_winfo.menu!=NULL) _winfo.menu->next=wmenu;
        wmenu->prev=_winfo.menu;
        wmenu->next=NULL;
        wmenu->parent=NULL;
        _winfo.menu=_winfo.cmenu=wmenu;
    }
    else {
        wmenu->parent=_winfo.cmenu;
        _winfo.cmenu=_winfo.cmenu->item->child=wmenu;
    }

    /* save info in menu record */
    wmenu->srow=srow;
    wmenu->scol=scol;
    wmenu->erow=erow;
    wmenu->ecol=ecol;
    wmenu->btype=btype;
    wmenu->battr=mapattr(battr);
    wmenu->wattr=mapattr(wattr);
    wmenu->open=open;
    wmenu->usecurr=NO;
    wmenu->item=NULL;

    /* increment menu level */
    _winfo.mlevel++;

    /* return normally */
    return(_winfo.errno=W_NOERROR);
}

/*---------------------------------------------------------------------------*/

int wmenuitem(int wrow,int wcol,char *str,int schar,int tagid,int fmask,
              void (*select)(void),unsigned hotkey,int help)
{
    struct _item_t *witem;

    /* make sure that wmenubeg() or wmenubegc() has been called */
    if(!_winfo.mlevel) return(_winfo.errno=W_NOMNUBEG);

    /* allocate memory for new item record */
    if((witem=malloc(sizeof(struct _item_t)))==NULL)
        return(_winfo.errno=W_ALLOCERR);

    /* add new menu item record to linked list */
    if(_winfo.cmenu->item!=NULL) _winfo.cmenu->item->next=witem;
    witem->prev=_winfo.cmenu->item;
    witem->next=NULL;
    _winfo.cmenu->item=witem;

    /* save info in item record */
    witem->wrow   = wrow;
    witem->wcol   = wcol;
    witem->str    = str;
    witem->schar  = schar;
    witem->tagid  = tagid;
    witem->fmask  = fmask;
    witem->select = select;
    witem->hotkey = hotkey;
    witem->desc   = NULL;
    witem->dwrow  = 0;
    witem->dwcol  = 0;
    witem->dattr  = 0;
    witem->redisp = NO;
    witem->help   = help;
    witem->child  = NULL;
    witem->before =
    witem->after  = NULL;

    /* set item level == menu level */
    _winfo.ilevel=_winfo.mlevel;

    /* return normally */
    return(_winfo.errno=W_NOERROR);
}

/*---------------------------------------------------------------------------*/

int wmenuend(int taginit,int menutype,int barwidth,int textpos,
             int textattr,int scharattr,int noselattr,int barattr)
{
    struct _item_t *item;
    int w_width,border,found;

    /* make sure at least 1 menu item has been defined */
    if(!_winfo.mlevel||_winfo.mlevel>_winfo.ilevel)
        return(_winfo.errno=W_NOITMDEF);

    /* make sure that the specified initial tagid matches the */
    /* tagid of one of the defined menu items in this menu    */
    for(found=NO,item=_winfo.cmenu->item;item!=NULL;item=item->prev) {
        if(item->tagid==taginit) {
            found=YES;
            break;
        }
    }
    if(!found) return(_winfo.errno=W_INVTAGID);

    /* if bar width > window width, then bar width = window width */
    border=(_winfo.cmenu->btype==5)?0:1;
    w_width=(_winfo.cmenu->ecol-border)-(_winfo.cmenu->scol+border)+1;
    if(barwidth>w_width) barwidth=w_width;

    /* add rest of menu information to menu record */
    _winfo.cmenu->citem     = NULL;
    _winfo.cmenu->tagcurr   = taginit;
    _winfo.cmenu->menutype  = menutype;
    _winfo.cmenu->barwidth  = barwidth;
    _winfo.cmenu->textpos   = barwidth?textpos:0;
    _winfo.cmenu->textattr  = mapattr(textattr);
    _winfo.cmenu->scharattr = mapattr(scharattr);
    _winfo.cmenu->noselattr = mapattr(noselattr);
    _winfo.cmenu->barattr   = _vinfo.mapattr?revsattr(_winfo.cmenu->textattr):
                              barattr;

    /* set current menu pointer to parent menu */
    if((_winfo.cmenu=_winfo.cmenu->parent)==NULL) _winfo.cmenu=_winfo.menu;

    /* decrement menu and item levels */
    _winfo.mlevel--;
    _winfo.ilevel--;

    /* return with no error */
    return(_winfo.errno=W_NOERROR);
}


/*---------------------------------------------------------------------------*/

int wmenuget(void)
{
    register unsigned int xch;
    register struct _item_t *citem;
    struct _item_t *item;
    int valid,pulldown=NO,menuerr,winerr,err;
    char ch;
    WINDOW w;

    /* make sure we have a defined menu */
    if(_winfo.cmenu==NULL) {
        _winfo.errno=W_NOMNUDEF;
        return(-1);
    }

    /* make sure that wmenuend() was called */
    if(_winfo.mlevel||_winfo.ilevel) {
        _winfo.errno=W_NOMNUEND;
        return(-1);
    }

    /* open menu's window (unless menu is to use current window) */
    if(!_winfo.cmenu->usecurr) {
        w=whandle();
        hide_mouse_cursor();
        if(!wopen(_winfo.cmenu->srow,_winfo.cmenu->scol,_winfo.cmenu->erow
          ,_winfo.cmenu->ecol,_winfo.cmenu->btype,_winfo.cmenu->battr
          ,_winfo.cmenu->wattr)) return(-1);
        show_mouse_cursor();

        /* save environment info, call menu open   */
        /* function, then restore environment info */
        if(_winfo.cmenu->open!=NULL) call_func(_winfo.cmenu->open);
    }

    /* if mouse cursor mode is on, show cursor */
    show_mouse_cursor();

    /* find first item in linked list */
    if((item=_winfo.cmenu->item)!=NULL) for(;item->prev!=NULL;item=item->prev);

    /* display all menu items */
    dispdesc=NO;
    valid=NO;
    while(item!=NULL) {
        disp_item(item,0);
        if(!(item->fmask&M_NOSEL)) valid=YES;
        item=item->next;
    }
    dispdesc=YES;

    /* make sure there is at least 1 selectable menu item */
    if(!valid) {
        pre_exit(w,YES);
        _winfo.errno=W_NOSELECT;
        return(-1);
    }

    /* search linked list of item records for the item matching    */
    /* the last item.  If there was no last item, search for       */
    /* the initial tag ID position.  If no initial tag ID position */
    /* was specified, then find the upper-leftmost menu item.      */
    citem=NULL;
    if(_winfo.cmenu->menutype&M_SAVE) {
        for(citem=_winfo.cmenu->item;citem!=NULL;citem=citem->prev) {
            if((_winfo.cmenu->citem==citem)&&
               (!(citem->fmask&M_NOSEL)))
                break;
        }
    }
    if(citem==NULL) {
        citem=wmenuifind(_winfo.cmenu->tagcurr);
        if((citem==NULL)||(citem->fmask&M_NOSEL)) citem=first_item();
    }

    /* call the current menu item's 'before'  */
    /* function and display current menu item */
    post_move(citem);

    /* main process of function */

    for(;;) {

        /* update current menu item and help category */
        _winfo.cmenu->citem=citem;
        _winfo.help=citem->help;

        /* read mouse/keyboard for keypress, then test the keypress */
        _kbinfo.inmenu=TRUE;
        xch=read_mouse(citem);
        citem=_winfo.cmenu->citem;
        if(!xch) xch=getxch();
        _kbinfo.inmenu=FALSE;
        switch(xch) {

            case 0x011b:    /* Esc */

                /* if Esc checking is on, erase selection bar, */
                /* free menu records and return to caller      */
                if(_winfo.esc||(_winfo.cmenu!=_winfo.menu)) {
                    pre_move(citem);
                    pre_exit(w,YES);
                    _winfo.errno=W_ESCPRESS;
                    return(-1);
                }
                break;

            case 0x4700:    /* Home */

                /* hide selection bar and find upper-leftmost menu item */
                citem=goto_item(citem,ITM_FR);
                break;

            case 0x4b00:    /* LeftArrow */

                /* if current menu is a pull-down menu,         */
                /* then erase selection bar, free menu records, */
                /* and return special code to caller            */
                if(_winfo.cmenu->menutype&M_PD) {
                    pre_move(citem);
                    pre_exit(w,YES);
                    _winfo.errno=W_NOERROR;
                    return(M_PREVPD);
                }

                /* if current menu is a horizontal menu, then hide */
                /* selection bar and find menu item to the left    */
                if(_winfo.cmenu->menutype&M_HORZ)
                    citem=goto_item(citem,ITM_LT);

                /* if pull-down menu flag is set, select this item */
                if(pulldown&&citem->child!=NULL) goto enter;
                break;

            case 0x4800:    /* UpArrow */

                /* if current menu is a vertical menu, then hide */
                /* selection bar and find menu item upwards      */
                if(_winfo.cmenu->menutype&(M_VERT|M_PD))
                    citem=goto_item(citem,ITM_UP);
                break;

            case 0x4d00:    /* RightArrow */

                /* if current menu is a pull-down menu,         */
                /* then erase selection bar, free menu records, */
                /* and return special code to caller            */
                if(_winfo.cmenu->menutype&M_PD) {
                    pre_move(citem);
                    pre_exit(w,YES);
                    _winfo.errno=W_NOERROR;
                    return(M_NEXTPD);
                }

                /* if current menu is a horizontal menu, then hide */
                /* selection bar and find menu item to the right   */
                if(_winfo.cmenu->menutype&M_HORZ)
                    citem=goto_item(citem,ITM_RT);

                /* if pulldown flag is set, then select this item */
                if(pulldown&&citem->child!=NULL) goto enter;
                break;

            case 0x5000:    /* DownArrow */

                /* if current item has a pull-down menu, select it */
                if(citem->fmask&M_HASPD) goto enter;

                /* if current menu is a vertical menu, then hide */
                /* selection bar and find menu item downwards    */
                if(_winfo.cmenu->menutype&(M_VERT|M_PD))
                    citem=goto_item(citem,ITM_DN);
                break;

            case 0x4f00:    /* End */

                /* hide selection bar and find */
                /* lower-rightmost menu item   */
                citem=goto_item(citem,ITM_LS);
                break;

            case 0x1c0d:    /* Enter */
enter:
                /* if current menu item has a pull-down menu */
                /* attached, then set the pulldown flag      */
                if(citem->fmask&M_HASPD) pulldown=YES;

                /* display menu item with selection bar */
                disp_item(citem,1);

                menuerr=0;

                /* if current menu item has a child menu */
                if(citem->child!=NULL) {

                    /* save environment info, process child */
                    /* menu, then restore environment info  */
                    _winfo.cmenu=citem->child;
                    hide_mouse_cursor();
                    err=whelpush();
                    menuerr=wmenuget();
                    winerr=_winfo.errno;
                    if(!err) whelpop();
                    show_mouse_cursor();
                    _winfo.cmenu=_winfo.cmenu->parent;

                    /* if an error was returned by   */
                    /* child menu, free menu records */
                    /* and pass error code to caller */
                    if(menuerr==-1&&winerr!=W_ESCPRESS) {
                        call_after(citem);
                        menuerr=winerr;
                        pre_exit(w,YES);
                        _winfo.errno=menuerr;
                        return(-1);
                    }
                }

                /* if the M_CLOSB feature is on, then close the menu's */
                /* window before the selection function is called.     */
                if(citem->fmask&M_CLOSB) close_window(w);

                /* this is used as a flag to see if the selection bar's */
                /* position has been changed by the select function.    */
                _winfo.cmenu->tagcurr=-1;

                /* if current menu item has a select function, call it */
                if(citem->select!=NULL) call_func(citem->select);

                /* if the M_CLOSB feature is on, then free the current */
                /* menu a before the selection function is called.     */
                if(citem->fmask&M_CLOSB) {
                    call_after(citem);
                    pre_exit(w,NO);
                    _winfo.errno=W_NOERROR;
                    return(citem->tagid);
                }

                /* check all menu items in current menu to  */
                /* see if their redisplay flag has been set */
                for(item=_winfo.cmenu->item;item!=NULL;item=item->prev) {
                    if(item->redisp) {
                        disp_item(item,(item==citem));
                        item->redisp=NO;
                    }
                }

                /* see if selection bar position was changed by select   */
                /* function.  if so, then move selection bar to new item */
                if(_winfo.cmenu->tagcurr!=-1) {
                    item=wmenuifind(_winfo.cmenu->tagcurr);
                    if(item!=NULL&&(!(item->fmask&M_NOSEL))) {
                        pre_move(citem);
                        post_move(_winfo.cmenu->citem=citem=item);
                        break;
                    }
                }

                /* if current menu item has a pull-down attached */
                if(citem->fmask&M_HASPD) {

                    /* if child menu returned previous pull-down menu */
                    /* flag, find menu item to the left and select it */
                    if(menuerr==M_PREVPD) {
                        citem=goto_item(citem,ITM_LT);
                        if(citem->fmask&M_HASPD) goto enter;
                        break;
                    }

                    /* if child menu returned next pull-down menu      */
                    /* flag, find menu item to the right and select it */
                   if(menuerr==M_NEXTPD) {
                        citem=goto_item(citem,ITM_RT);
                        if(citem->fmask&M_HASPD) goto enter;
                        break;
                    }
                }

                /* turn off pulldown flag */
                pulldown=NO;

                /* if child menu returned an exit-all-menus flag, */
                /* then free menu records and pass exit-all-menus */
                /* flag onto caller                               */
                if((menuerr==M_EXITALL)||(citem->fmask&M_CLALL)) {
                    disp_item(citem,1);
                    call_after(citem);
                    pre_exit(w,YES);
                    _winfo.errno=W_NOERROR;
                    return(M_EXITALL);
                }

                /* unless an exit-this-menu flag was returned by the */
                /* child menu, or current item has close-menu        */
                /* specified, free menu records, and return tag      */
                /* identifier of current menu item                   */
                if(citem->child!=NULL||citem->select!=NULL)
                    if((menuerr!=M_EXIT)&&(!(citem->fmask&M_CLOSE)))
                        break;
                call_after(citem);
                pre_exit(w,YES);
                _winfo.errno=W_NOERROR;
                return(citem->tagid);

            default:

                /* separate ASCII code from keypress code, if ASCII */
                /* code is zero, then it must be an extended key    */
                ch=(char)xch;
                if(!ch) break;

                /* scan through list of items for one that */
                /* has a tag identifier matching keypress  */
                valid=YES;
                item=citem->next;
                for(;;) {
                    while(item!=NULL) {
                        if((toupper(ch)==toupper(item->schar))&&(!(item->fmask
                          &M_NOSEL))) goto farbreak;
                        if(citem==item) {
                            valid=NO;
                            goto farbreak;
                        }
                        item=item->next;
                    }
                    for(item=_winfo.cmenu->item;item->prev!=NULL;
                      item=item->prev);
                }
farbreak:

                /* if a matching tag identifier was found,          */
                /* then hide selection bar, and if quick-key        */
                /* selection is allowed, select the found menu item */
                if(valid) {
                    if(item!=citem) {
                        pre_move(citem);
                        post_move(_winfo.cmenu->citem=citem=item);
                    }
                    if(!(_winfo.cmenu->menutype&M_NOQS)) goto enter;
                }
        }
    }
}

/*---------------------------------------------------------------------------*/

/* this function will calculate the width of the selection bar */

static int calc_bar_width(struct _menu_t *wmenu,struct _item_t *witem)
{
    register int width;

    width=strlen(witem->str);
    if(wmenu->barwidth) width=wmenu->barwidth;
    return(width);
}

/*---------------------------------------------------------------------------*/

/* this function will calculate the window column */
/* at the center of the given menu item           */

static int calc_center_item(struct _item_t *item)
{
    return( ((int)item->wcol) + (strlen(item->str)/2) );
}

/*---------------------------------------------------------------------------*/

/* this function will call a menu item's "after" function if it exists */

static void call_after(struct _item_t *citem)
{
    if(citem->after!=NULL) call_func(citem->after);
}

/*---------------------------------------------------------------------------*/

/* this function will call a menu item's "before" function if it exists */

static void call_before(struct _item_t *citem)
{
    if(citem->before!=NULL) call_func(citem->before);
}

/*---------------------------------------------------------------------------*/

/* this function calls the given function */

static void call_func(void (*func)(void))
{
    struct _menu_t *menu;
    WINDOW w;
    int err;

    hide_mouse_cursor();
    menu=_winfo.cmenu;
    w=whandle();
    err=whelpush();
    (*func)();
    wactiv(w);
    if(!err) whelpop();
    _winfo.cmenu=menu;
    show_mouse_cursor();
}

/*---------------------------------------------------------------------------*/

/* this function closes the current menu's window and reactivates */
/* the menu window open prior to opening the current menu window  */

static void close_window(WINDOW w)
{
    if(!_winfo.cmenu->usecurr) {
        hide_mouse_cursor();
        wclose();
        wactiv(w);
        show_mouse_cursor();
    }
}

/*---------------------------------------------------------------------------*/

/* this function displays a menu selection, using selection bar if specified */

static void disp_item(struct _item_t *witem,int bar)
{
    register char *p;
    register int i;
    int ch,textend,width,chattr,wcol,found=NO;

    /* if mouse cursor is on, temporarily hide it */
    hide_mouse_cursor();

    /* initialize width of output and end of text */
    p=witem->str;
    width=calc_bar_width(_winfo.cmenu,witem);
    textend=_winfo.cmenu->textpos+strlen(p)-1;
    wgotoxy(witem->wrow,wcol=witem->wcol);

    /* display menu item including selection bar */
    for(i=0;i<width;i++) {

        /* see if currently in bar region.  if so, then use  */
        /* a space for the character. otherwise use the      */
        /* character from the current position in the string */
        ch=(i<_winfo.cmenu->textpos||i>textend)?' ':(*p++);

        /* select attribute of character to be displayed based upon if      */
        /* selection bar was specified, if the menu item is non-selectable, */
        /* if the character is a tag character, or none of the above.       */
        if(bar)
            chattr=_winfo.cmenu->barattr;
        else if(witem->fmask&M_NOSEL)
            chattr=_winfo.cmenu->noselattr;
        else if((ch==witem->schar)&&!found) {
            found=YES;
            chattr=_winfo.cmenu->scharattr;
        }
        else
            chattr=_winfo.cmenu->textattr;

        /* display character in selected attribute */
        wprintc(witem->wrow,wcol++,chattr,ch);
    }

    /* display text description, if one exists */
    if(witem->desc!=NULL&&dispdesc) {
        wgotoxy(witem->dwrow,witem->dwcol);
        wtextattr(witem->dattr);
        wputs(witem->desc);
        wclreol();
    }

    /* if mouse cursor is hidden, unhide it */
    show_mouse_cursor();
}

/*---------------------------------------------------------------------------*/

/* this function finds the next menu selection down */

static struct _item_t *down_item(struct _item_t *curr)
{
    struct _item_t *best,*temp;
    int brow,bcol,tcol,trow,crow,ccol,tdist,bdist;

    /* initialize best record to NULL */
    best=NULL;
    brow=bcol=32767;

    /* calculate window column at center of current item */
    crow=(int)curr->wrow;
    ccol=calc_center_item(curr);

    /* search backwards through linked list, testing each item */
    for(temp=_winfo.cmenu->item;temp!=NULL;temp=temp->prev) {

        /* calculate window column at center of test item */
        trow=(int)temp->wrow;
        tcol=calc_center_item(temp);

        if(trow>crow) {
            tdist=abs((ccol-tcol));
            bdist=abs((ccol-bcol));
            if((trow<brow)||((trow==brow&&tdist<bdist))) {
                best=temp;
                brow=trow;
                bcol=tcol;
            }
        }
    }

    /* if there wasn't an item downwards, then wrap around */
    if(best==NULL) {
        if((temp=malloc(sizeof(struct _item_t)))==NULL) {
            best=curr;
        }
        else {
            *temp=*curr;
            temp->wrow=-1;
            best=down_item(temp);
            free(temp);
        }
    }
    else
        /* see if menu selection is non-selectable */
        if(best->fmask&M_NOSEL)
            best=down_item(best);

    /* return best record */
    return(best);
}

/*---------------------------------------------------------------------------*/

/* this function finds the upper-leftmost menu selection */

static struct _item_t *first_item(void)
{
    struct _item_t *best,*temp;

    /* initialize best record to highest record in linked list */
    best=_winfo.cmenu->item;

    /* search backwards through linked list, testing each item */
    for(temp=best->prev;temp!=NULL;temp=temp->prev) {
        if( (temp->wrow<best->wrow) ||
            ( (temp->wrow==best->wrow) && (temp->wcol<best->wcol) ) )
                best=temp;
    }

    /* see if menu selection is non-selectable */
    if(best->fmask&M_NOSEL) best=right_item(best);

    /* return best record */
    return(best);
}

/*---------------------------------------------------------------------------*/

/* this is a recursive function that frees a menu and all of its submenus */

static void free_menu(struct _menu_t *wmenu)
{
    struct _item_t *witem;

    /* free all items in menu, including sub-menus */
    while(wmenu->item!=NULL) {
        if(wmenu->item->child!=NULL) free_menu(wmenu->item->child);
        witem=wmenu->item->prev;
        free(wmenu->item);
        wmenu->item=witem;
        if(wmenu->item!=NULL) wmenu->item->next=NULL;
    }

    /* free the menu itself */
    free(wmenu);
}
 
/*---------------------------------------------------------------------------*/

/* this function will hide the mouse cursor */
/* if mouse cursor mode is on               */

static void hide_mouse_cursor(void)
{
    if(_mouse&MS_CURS) mshidecur();
}

/*---------------------------------------------------------------------------*/

/* this function moves the selection bar to another menu item,  */
/* automatically taking care of necessary pre/post move actions */

static struct _item_t *goto_item(struct _item_t *citem,int which)
{
    struct _item_t *item;

    if(which==ITM_FR)       item=first_item();
    else if(which==ITM_LS)  item=last_item();
    else                    item=(*funcs[which])(citem);
    if(item!=citem) {
        pre_move(citem);
        post_move(citem=item);
    }
    return(citem);
}

/*---------------------------------------------------------------------------*/

/* this function will find the bottom-rightmost menu selection */

static struct _item_t *last_item(void)
{
    struct _item_t *best,*temp;
    int bcol;

    /* initialize best record to highest record in linked list */
    best=_winfo.cmenu->item;
    bcol=best->wcol;

    /* search backwards through linked list, testing each item */
    for(temp=best->prev;temp!=NULL;temp=temp->prev) {
        if( (temp->wrow>best->wrow) ||
            ( (temp->wrow==best->wrow) &&
                (temp->wcol>bcol) ) ) {
                    best=temp;
                    bcol=best->wcol;
        }
    }

    /* see if menu selection is non-selectable */
    if(best->fmask&M_NOSEL) best=left_item(best);

    /* return best record */
    return(best);
}

/*---------------------------------------------------------------------------*/

/* this function will find the menu selection to the left of current */

static struct _item_t *left_item(struct _item_t *curr)
{
    struct _item_t *best,*temp;
    int wwidth,bpos,tpos,cpos;

    /* calculate window width and current position */
    wwidth=_winfo.cmenu->ecol - _winfo.cmenu->scol + 1;
    cpos=(curr->wrow * wwidth) + curr->wcol;

    /* initialize best record to NULL, and best position to -1 */
    best=NULL;
    bpos=-1;

    /* search backwards through linked list, testing each item */
    for(temp=_winfo.cmenu->item;temp!=NULL;temp=temp->prev) {

        /* calculate position of test each item */
        tpos=(temp->wrow*wwidth) + temp->wcol;

        /* compare position of test item with best item, current item */
        if(tpos>bpos && tpos<cpos) {
            best=temp;
            bpos=tpos;
        }
    }

    /* if there wasn't a item to the left, then wrap around */
    if(best==NULL)
        best=last_item();
    else
        /* see if menu selection is non-selectable */
        if(best->fmask&M_NOSEL)
            best=left_item(best);

    /* return best record */
    return(best);
}

/*---------------------------------------------------------------------------*/

/* this function determines if the mouse cursor is on a menu item */

static struct _item_t *mouse_on_item(struct _menu_t *menu,int mcrow,int mccol)
{
    int srow,scol,border,start,end;
    struct _item_t *item,*found;

    found  = NULL;
    srow   = menu->srow;
    scol   = menu->scol;
    border = menu->btype==5?0:1;
    for(item=menu->item;item!=NULL;item=item->prev) {
        if(mcrow==(srow+border+item->wrow)) {
            start = scol+border+item->wcol;
            end   = start+calc_bar_width(menu,item)-1;
            if(mccol>=start&&mccol<=end) {
                found=item;
                break;
            }
        }
    }
    return(found);
}

/*---------------------------------------------------------------------------*/

/* this function is called after a menu bar move */

static void post_move(struct _item_t *citem)
{
    _winfo.cmenu->citem=citem;
    _winfo.help=citem->help;
    disp_item(citem,1);
    call_before(citem);
}

/*---------------------------------------------------------------------------*/

static void pre_exit(WINDOW w,int close)
{
    struct _menu_t *wmenu;

    hide_mouse_cursor();

    /* if not using current window for menu, then close it */
    if(close) close_window(w);

    /* if at highest menu then free the whole menu structure */
    if(_winfo.cmenu==_winfo.menu) {
        wmenu=_winfo.menu->prev;
        if(_winfo.cmenu!=NULL) free_menu(_winfo.cmenu);
        _winfo.menu=wmenu;
        if(_winfo.menu!=NULL) _winfo.menu->next=NULL;
        _winfo.cmenu=_winfo.menu;
    }
}

/*---------------------------------------------------------------------------*/

/* this function prepares for a menu bar move */

static void pre_move(struct _item_t *citem)
{
    disp_item(citem,0);
    call_after(citem);
}

/*---------------------------------------------------------------------------*/

/* this function reads the mouse for input */

static int read_mouse(struct _item_t *citem)
{
    register struct _item_t *item;
    int bcount,bstat,mrow,mcol;

    /* if free-floating mouse cursor support is on */
    if(_mouse&MS_CURS) {

        /* clear mouse button queue */
        msbclear();

        /* loop until a key is pressed */
        while(!kbhit()&&_kbinfo.kbuf==NULL) {

            /* call the keyboard loop function */
            if(_kbinfo.kbloop!=NULL) (*_kbinfo.kbloop)();

            /* if left button was pressed, and mouse cursor is on  */
            /* a selectable menu item, then move selection bar to  */
            /* that item, and select it.  If mouse cursor is on    */
            /* a main menu item of a pull-down menu system, then   */
            /* stuff that item's selection character into the CXL  */
            /* keyboard buffer and return Esc to close the current */
            /* pull-down menu.                                     */
            msbreles(0,&bstat,&bcount,&mrow,&mcol);
            if(bcount) {
                if((item=mouse_on_item(_winfo.cmenu,mrow,mcol))==NULL) {
                    if(_winfo.cmenu->menutype&M_PD) {
                        if(((item=mouse_on_item(_winfo.cmenu->parent,mrow,
                           mcol))!=NULL)&&(!(item->fmask&M_NOSEL))) {
                            kbput(item->schar);
                            return(0x011b);     /* Esc */
                        }
                    }
                }
                else {
                    if(!(item->fmask&M_NOSEL)) {
                        if(citem!=item) {
                            pre_move(citem);
                            post_move(_winfo.cmenu->citem=citem=item);
                        }
                        return(0x1c0d);         /* Enter */
                    }
                }
            }

            /* if right button was pressed, simulate pressing the Esc key */
            msbreles(1,&bstat,&bcount,&mrow,&mcol);
            if(bcount) return(0x011b);          /* Esc */
        }
    }

    /* return zero - it means a key was pressed */
    return(0);
}

/*---------------------------------------------------------------------------*/

/* this function will find the menu selection to the right of current */

static struct _item_t *right_item(struct _item_t *curr)
{
    struct _item_t *best,*temp;
    int wwidth,bpos,tpos,cpos;

    /* calculate window width and current position */
    wwidth=_winfo.cmenu->ecol - _winfo.cmenu->scol + 1;
    cpos=(curr->wrow * wwidth) + curr->wcol;

    /* initialize best record to NULL, and best position to 32767 */
    best=NULL;
    bpos=32767;

    /* search backwards through linked list, testing items */
    for(temp=_winfo.cmenu->item;temp!=NULL;temp=temp->prev) {

        /* calculate position of test item */
        tpos=(temp->wrow*wwidth) + temp->wcol;

        /* compare position of test item with best item, current item */
        if(tpos<bpos && tpos>cpos) {
            best=temp;
            bpos=tpos;
        }
    }

    /* if there wasn't a item to the right, then wrap around */
    if(best==NULL)
        best=first_item();
    else
        /* see if menu selection is non-selectable */
        if(best->fmask&M_NOSEL)
            best=right_item(best);

    /* return best record */
    return(best);
}

/*---------------------------------------------------------------------------*/

/* this function will display the mouse */
/* cursor if mouse cursor mode is on    */

static void show_mouse_cursor(void)
{
    if(_mouse) {
        if(_mouse&MS_CURS) {
            msshowcur();
            mscursor(0,0xffff,((LGREY|_LGREY)<<8));
        }
    }
}

/*---------------------------------------------------------------------------*/

/* this function finds the previous menu selection upwards */

static struct _item_t *up_item(struct _item_t *curr)
{
    struct _item_t *best,*temp;
    int brow,bcol,tcol,crow,trow,ccol,tdist,bdist;

    /* initialize best record to NULL */
    best = NULL;
    brow = -1;
    bcol = 32767;

    /* calculate window column at center of current item */
    crow=(int)curr->wrow;
    ccol=calc_center_item(curr);

    /* search backwards through linked list, testing items */
    for(temp=_winfo.cmenu->item;temp!=NULL;temp=temp->prev) {

        /* calculate window column at center of test item */
        trow=(int)temp->wrow;
        tcol=calc_center_item(temp);

        if(trow<crow) {
            tdist=abs(ccol-tcol);
            bdist=abs(ccol-bcol);
            if((trow>brow)||((trow==brow&&tdist<bdist))) {
                best=temp;
                brow=trow;
                bcol=tcol;
            }
        }
    }

    /* if there wasn't a item to the left, then wrap around */
    if(best==NULL) {
        if((temp=malloc(sizeof(struct _item_t)))==NULL) {
            best=curr;
        }
        else {
            *temp=*curr;
            temp->wrow=255;
            best=up_item(temp);
            free(temp);
        }
    }
    else
        /* see if menu selection is non-selectable */
        if(best->fmask&M_NOSEL)
            best=up_item(best);

    /* return best record */
    return(best);
}

