/*  CXL - Copyright (c) 1987-1989 by Mike Smedley - All Rights Reserved  */
/*  WMOVE.C     - moves active window to a new location     */

#include "cxlvid.h"
#include "cxlwin.h"

int wmove(int nsrow,int nscol)
{
    int *p,*q,nerow,necol,shattr=-1;

    /* check for active windows */
    if(!_winfo.total) return(_winfo.errno=W_NOACTIVE);

    /* check for valid coordinates */
    if(nsrow<0||nscol<0) return(_winfo.errno=W_INVCOORD);

    /* if window has shadow, close it before the move */
    if(_winfo.active->wsbuf!=NULL) {
        shattr=_winfo.active->wsattr;
        wshadoff();
    }

    /* save current window */
    if((q=wsave(_winfo.active->srow,_winfo.active->scol,_winfo.active->erow,
       _winfo.active->ecol))==NULL) return(_winfo.errno=W_ALLOCERR);

    /* restore what current window's buffer holds */
    wrestore(_winfo.active->wbuf);

    /* calculate new end row and new end column */
    nerow=_winfo.active->erow-_winfo.active->srow+nsrow;
    necol=_winfo.active->ecol-_winfo.active->scol+nscol;

    /* save area of screen where window will be moved to */
    if((p=wsave(nsrow,nscol,nerow,necol))==NULL)
        return(_winfo.errno=W_ALLOCERR);

    /* change coordinates in saved window's buffer
       and restore window to new coordinates */
    *(q)   = nsrow;
    *(q+1) = nscol;
    *(q+2) = nerow;
    *(q+3) = necol;
    wrestore(q);

    /* update window record */
    _winfo.active->wbuf   = p;
    _winfo.active->row    = _winfo.active->row-_winfo.active->srow+nsrow;
    _winfo.active->column = _winfo.active->column-_winfo.active->scol+nscol;
    _winfo.active->srow   = nsrow;
    _winfo.active->scol   = nscol;
    _winfo.active->erow   = nerow;
    _winfo.active->ecol   = necol;

    /* if window has shadow, redraw it */
    if(shattr!=-1) wshadow(shattr);

    /* reset cursor position */
    gotoxy_(_winfo.active->row,_winfo.active->column);

    /* return normally */
    return(_winfo.errno=W_NOERROR);
}
