/*  CXL - Copyright (c) 1987-1989 by Mike Smedley - All Rights Reserved  */
/*  WOPEN.C  -  wopen() opens a window and makes it active
                wclose() closes a window                    */

#include <stdlib.h>
#include "cxldef.h"
#include "cxlvid.h"
#include "cxlwin.h"

WINDOW wopen(int srow,int scol,int erow,int ecol,int btype,int battr,int wattr)
{
    int *wbuf;
    struct _wrec_t *wrec;
    register int border;

    /* check for valid box type */
    if(btype<0 || btype>5) {
        _winfo.errno=W_INVBTYPE;
        return(0);
    }

    /* see if window is to have a border */
    border=((btype==5)?NO:YES);

    /* check for valid coordinates */
    if(srow>(erow-border) || scol>(ecol-border)) {
        _winfo.errno=W_INVCOORD;
        return(0);
    }

    /* check for monochrome adapter, adjust attributes */
    battr=mapattr(battr);
    wattr=mapattr(wattr);

    /* allocate memory for new record */
    wrec=malloc(sizeof(struct _wrec_t));
    if(wrec==NULL) {
        _winfo.errno=W_ALLOCERR;
        return(0);
    }

    /* save affected area of screen */
    if((wbuf=wsave(srow,scol,erow,ecol))==NULL) {
        free(wrec);
        _winfo.errno=W_ALLOCERR;
        return(0);
    }

    /* add new record to linked list */
    if(_winfo.active!=NULL) _winfo.active->next=wrec;
    wrec->prev=_winfo.active;
    wrec->next=NULL;
    _winfo.active=wrec;

    /* draw and fill text box on screen */
    if(border) box_(srow,scol,erow,ecol,btype,battr);
    fill_(srow+border,scol+border,erow-border,ecol-border,_winfo.fillch,wattr);

    /* increment window handle counter */
    _winfo.handle++;

    /* save window info in window record */
    _winfo.active->wbuf=wbuf;
    _winfo.active->whandle=_winfo.handle;
    _winfo.active->srow=srow;
    _winfo.active->scol=scol;
    _winfo.active->erow=erow;
    _winfo.active->ecol=ecol;
    _winfo.active->btype=btype;
    _winfo.active->wattr=wattr;
    _winfo.active->battr=battr;
    _winfo.active->border=border;
    _winfo.active->row=srow+border;
    _winfo.active->column=scol+border;
    _winfo.active->attr=wattr;
    _winfo.active->title=NULL;
    _winfo.active->tpos=
    _winfo.active->help=0;
    _winfo.active->form=NULL;
    _winfo.active->wsbuf=NULL;

    /* increment total number of open windows */
    _winfo.total++;

    /* initialize cursor location to window row 0 column 0 */
    wgotoxy(0,0);

    /* return normally */
    _winfo.errno=W_NOERROR;
    return(_winfo.handle);
}

/*---------------------------------------------------------------------------*/

int wclose(void)
{
    struct _wrec_t *wrec;

    /* check for active window */
    if(!_winfo.total) return(_winfo.errno=W_NOACTIVE);

    /* if window has a shadow, close shadow first */
    if(_winfo.active->wsbuf!=NULL) wshadoff();

    /* restore contents of and free memory held by window */
    wrestore(_winfo.active->wbuf);

    /* decrement total number of open windows */
    _winfo.total--;

    /* free memory held by window's record and update linked list */
    wrec=_winfo.active->prev;
    free(_winfo.active);
    _winfo.active=wrec;
    if(_winfo.active!=NULL) _winfo.active->next=NULL;

    /* update cursor location and help category */
    if(_winfo.active!=NULL) {
        gotoxy_(_winfo.active->row,_winfo.active->column);
        if(_winfo.active->help) _winfo.help=_winfo.active->help;
    }

    /* return normally */
    return(_winfo.errno=W_NOERROR);
}

